/* interface to 32 kHz timer */
/* Copyright (C) 2009 Thomas Klosa
 * Copyright (C) 2010 Thomas Klosa, Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file
 * @ingroup geXPal_lib
 * @brief  32kHz Timer Implementation
 * @todo Add function to register timer callback functions for application layer
 */


#include <avr/io.h>
#include <avr/interrupt.h>   
#include <util/delay.h>

#include "keys.h"
#include "power.h"
#include "async32khz.h"


#define HL_ASYN_FAST_INC 5  /** number of async timer ticks for one 
							 * fast system timer tick
							 */
/** @brief Async timer overflow increment storage
 */
uint8_t FastTimerInc;

/** @brief First element in slow timer callback list
 */
cb_list_t cbListSlowFirst;

/** @brief First element in fast timer callback list
 */
cb_list_t cbListFastFirst;

/** @brief System Timer Call Status
 *
 *  This bitmap variable will mirror the current timer call and Interrupt status
 *  It will allow the recognition of overun effects and lock the recursive call of timer functions.
 */
uint8_t SysTimerStatus;

/** @brief Init async timer system
 */
void hl_asyn_init(void){

	// init callback lists
#if 0 /* not needed since data gets initialised to 0 by startup */
	cbListSlowFirst = NULL;

	cbListFastFirst = NULL;
#endif

	// init Status
	SysTimerStatus = ASYNC_IDLE;

	// run timer from 32kHz crystal and generate 
	// 1 interrupt p. sec.
	ASSR=0x20; 

	// allow xtal to startup
	_delay_ms(1000);

	TCCR2A=0x00;	// use in normal mode, disable OCpins
	TCCR2B=0x05;	// normal mode, clkSource = T2S/128

	FastTimerInc = HL_ASYN_FAST_INC;
	OCR2A = FastTimerInc;	// Overrun value for fast system timer

	// wait until all busy flags are clear
	while ( ASSR & 0x1f );


	// set overrun interrupt for 1sec tick
	// and CompareMatchA for ~20ms fast system tick
	TIFR2 |= _BV(OCF2B) | _BV(OCF2A) | _BV(TOV2);

	// wait until all busy flags are clear
	while ( ASSR & 0x1f );

	TIMSK2 = _BV(OCIE2A) | _BV(TOIE2);

}


/** @brief Overflow used as 1sec, slow system timer
*/
ISR(TIMER2_OVF_vect)
{
	TIFR2 |= _BV(OCF2A);
	TIMSK2 = _BV(OCIE2A) | _BV(TOIE2);

	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_1S_TMR);

	if( (SysTimerStatus & ASYNC_SLOW_ACTIVE) == 0 ){
	
		// no slow int execution active, activate
		SysTimerStatus &= ASYNC_SLOW_ACTIVE;
	
		if( (SysTimerStatus & ASYNC_FAST_ACTIVE) == 0 ){
	
			// Fast timer is not active, enable other interrupts as long 
			// as slow functions get called
			sei();					
			callback_iterate (cbListSlowFirst);
			cli();	
		}else{
			// Fast timer is active, set block flag
			SysTimerStatus &= ASYNC_SLOW_BLOCK;
			// The fast timer will call the slow timer execution when done
			// This will avoid recursive calls					
		}
	}else{
		// slow int still in execution, set overrun
		SysTimerStatus &= ASYNC_SLOW_OVERRUN;	
	}
}


/** @brief 19,53ms timer
 * 
 * Compare value is incremented by HL_ASYN_FAST_INC during each ISR call
 */
ISR(TIMER2_COMPA_vect)
{
	
	FastTimerInc += HL_ASYN_FAST_INC;

	// on overun, stop int until overrun
	if (FastTimerInc < HL_ASYN_FAST_INC){
		TIMSK2 = _BV(TOIE2);
	}
	
	OCR2A = FastTimerInc;	// Compare value for next fast system timer

	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_SYS_TMR);

	if( (SysTimerStatus & ASYNC_FAST_ACTIVE) == 0 ){
	
		// no fast int execution active, activate
		SysTimerStatus &= ASYNC_FAST_ACTIVE;
	
		if( (SysTimerStatus & ASYNC_SLOW_ACTIVE) == 0 ){
	
			// Slow timer is not active, enable other interrupts as long 
			// as slow functions get called
			sei();					
			callback_iterate (cbListFastFirst);
			cli();	
		}else{
			// Slow timer is active, set block flag
			SysTimerStatus &= ASYNC_FAST_BLOCK;
			// The fast timer will call the slow timer execution when done
			// This will avoid recursive calls					
		}
	}else{
		// Fast int still in execution, set overrun
		SysTimerStatus &= ASYNC_FAST_OVERRUN;	
	}
}


/** @brief Register a call back function for the slow timer
 *
 *	The function will add an entry to the list of functions to be called when the 1sec timer expires
 *
 * The new entry is added to the front of the list.
 *  
 */
 void hl_asyn_Register1sTimer(cb_function_t fkt2call, void *arg)
{
  cb_register_call (&cbListSlowFirst, fkt2call, arg);
}

/** @brief Register a call back function for the fast timer
 *
 *	The function will add an entry to the list of functions to be called when the 19,53ms timer expires
 *
 * The new entry is added to the front of the list.
 *  
 */
 void hl_asyn_RegisterFastTimer(cb_function_t fkt2call, void *arg)
{
  cb_register_call (&cbListFastFirst, fkt2call, arg);
}

