/* template header file */
/* Copyright (C) 2009 Thomas Klosa
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file
 * @ingroup geXPal_lib
 * @brief Header for the async Timer Base Functions
 * @todo Add function to register timer callback functions for application layer
 */


#ifndef ASYNC32KHZ_H
#define ASYNC32KHZ_H

#include "cb_list.h"

/** @brief Idle, no timer Interrupt pending
 */
#define ASYNC_IDLE	0x00

/** @brief Slow INT active
 */
#define ASYNC_SLOW_ACTIVE	0x01

/** @brief Fast INT pending
 */
#define ASYNC_FAST_ACTIVE	0x02

/** @brief Slow INT overrun
 */
#define ASYNC_SLOW_OVERRUN	0x10

/** @brief Fast INT overrun
 */
#define ASYNC_FAST_OVERRUN	0x20

/** @brief Fast INT blocked by Slow INT
 *
 *  A fast interrupt was called while the slow INT was in execution
 *  At the end of the slow ISR, the fast int will get called
 */
#define ASYNC_FAST_BLOCK	0x40

/** @brief SLOW INT blocked by FAST INT
 *
 *  A slow interrupt was called while the fast INT was in execution
 *  At the end of the fast ISR, the slow int will get called
 */
#define ASYNC_SLOW_BLOCK	0x80

/**
 * @brief Initialize the 32kHz timer (timer2)
 *
 * - 1s overrun
 * - OverflowA to generate 19.53ms interrupts
 * 
 * 
 */
void hl_asyn_init(void);

/**
 * @brief Start the system Timer again
 *
 * Restart the system timer again when it was stopped to save power 
 */
inline void hl_asyn_StartSysTimer(void){
	/* Start System Timer if not running */
	TIMSK2 |= _BV(OCIE2A);
}

/**
 * @brief Stop the system Timer
 *
 * Stop the system timer to save power 
 */
inline void hl_asyn_StopSysTimer(void){
	/* Start System Timer if not running */
	TIMSK2 &= ~(_BV(OCIE2A));
}


/** @brief Register a call back function for the slow timer
 *
 *	The function will add an entry to the list of functions to be called when the 1sec timer expires
 *
 * The new entry is added to the front of the list.
 *  
 */
void hl_asyn_Register1sTimer(cb_function_t fkt2call, void *arg);


/** @brief Register a call back function for the fast timer
 *
 *	The function will add an entry to the list of functions to be called when the 19,53ms timer expires
 *
 * The new entry is added to the front of the list.
 *  
 */
 void hl_asyn_RegisterFastTimer(cb_function_t fkt2call, void *arg);



#endif
