/* Project eXplorersPal
   Power Management Module for Prototype A
 */
/* Copyright (C) 2009 Thomas Klosa, Holger Dietze
                 2010 Thomas Klosa
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file
 * @ingroup geXPal_lib
 * @brief  Power Management Routines for the eXPal Prototype A

\par This module contains routines for the following functions:

- Initialize the hardware in a clear way to ensure clean logic levels on all lines. Floating signals would cause extra power consumtion.
- Switching functions for the power hardware
- Power State management and sleep
- Battery Voltage control and capasity estimation

\par Basic Concept for the power management

	The application has to send the system in a power save state by calling
	hl_pwr_EnterSleep();
	This function does return the wake-up source. Based on that the application can
	determine if there is something to do or if this function can be called again 
	immediately.

	The individual application modules can set the allowed sleep state by calling
	hl_pwr_CtrlAutoSave();

	When HL_PWR_ALLOW_OFF is set, the system will turn into this state when
	the timer is down to zero and USB is not active. In this case the 3.3V
	will get switched off.
	The timer can be set to zero by calling hl_pwr_OffNow();

	By calling hl_pwr_AppBusy(); the timer is set to HL_PWR_OFF_TIME again.
*/

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>   
#include <util/delay.h>

#include "expal_board.h"
#include "power.h"
#include "async32khz.h"
#include "adc.h"
#include "cb_list.h"

/*
	Variables
*/
volatile hl_pwr_state_t hl_pwr_state;	// System power state (current setting)

volatile hl_pwr_state_t hl_pwr_info;		// Power States,the system is allowed to use

volatile uint16_t	PowerOffTimer;

volatile uint16_t	BatVoltage;	// voltage format: n.f	(n << 8 | (uint8_t)(f*256.0))

/** @brief First element in power callback list
 */
cb_list_t cbListPowerFirst;

/** @brief Bitmap variable to indicate the wake-up source
 *
 *	WakeUpSource is a symmary of system actions that caused the system to wake up.
 *	It's cleared before entering sleep, any wake-up ISR will set the individual bit.
 *	When checked some time after wake-up, more than one source might be active
 */
uint8_t	volatile WakeUpSource;

/** @brief Primary Hardware Init Function
 *
 *	This function has to be called as the first action in main()
 *  It will ensure the correct setting of all processor IO lines and other board hardware.
 *	The function will only set everything in a defined, power saving state and not init any hardware unit to work.
 */
void hl_pwr_InitBoard(void){

	/*
		Initialize the ports to avoid floating signals and excessive power
		consumption.
		To be complete it is more obvious to access the ports directly here.
	*/
	// Port A
	// can stay as is after reset

	// Port B
	PORTB= 0x80 | _BV(HL_SD_SS);  // USB_RESET inakt, sauberer Pegel fr SD Treiber
	DDRB = 0x87; 

	// Port C
	DDRC = 0xFF;  // Addr.Augaenge
	PORTC= 0xC0;  // RAM Disable

	// Port D
	PORTD = 0xFF; // PullUp on for KeyLines

	// Port E
	PORTE = 0xFF; // PullUp on

	// Port F
	PORTF = 0x0F & ~_BV(PF0); // PullUp for unused AD Inputs

	// Port G
	DDRG = 0x07;
	PORTG = 0x07; // RD/WR disable

	// Port H
	PORTH = 0xFF; // PullUp on

	// Port J
	PORTJ = 0xFF; // PullUp on

	// Port K
	PORTK = 0xFF; // PullUp on

	// Port L
	DDRL = 0x07 | _BV(HL_EX232_FRCON) | _BV(HL_EX232_EN);

	// 5V ON, pull up at open pins
	PORTL = _BV(HL_PWR_VCC5) | _BV(PL3)|_BV(PL4)
	  | 0*_BV(HL_EX232_FRCON) | 1*_BV(HL_EX232_EN);

	/* set power reduction registers */
	PRR0 = _BV(PRTWI) /* | _BV(PRTIM2) */ | _BV(PRTIM0)
		| _BV(PRTIM1) | _BV(PRSPI) | _BV(PRUSART0)
	    | _BV(PRADC)
	    ;
	PRR1 = _BV(PRTIM5) | _BV(PRTIM4) | _BV(PRTIM3)
	  	| _BV(PRUSART3) | _BV(PRUSART2) | _BV(PRUSART1)
	    ;

}

// function prototypes
void hl_pwr_slow(void*);
void hl_pwr_fast(void*);

/** @brief Init power management system
 *
 * Initialize
 * - Variables
 * - Bat Voltage measurement
 * - Key interrupts
 */

void hl_pwr_InitPwrMan(void){

	hl_pwr_state = HL_PWR_5V_ACTIVE | HL_PWR_ST_RUNNING;
	hl_pwr_info = 0; // Stay in full on
	PowerOffTimer = HL_PWR_OFF_TIME;
	BatVoltage = HL_PWR_VOLT_INVALID | HL_PWR_MEASURE_ACTIVE;

	// Initiate first BatVoltage measurement
	hl_adc_init();
	hl_adc_start ();
	// Measure flag already set in BatVoltage var init

	// Init key interrupts
	EICRA = 0x00;  // all low level ints
	EIMSK |= (_BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3));

	hl_asyn_Register1sTimer( (&hl_pwr_slow), NULL);
	hl_asyn_RegisterFastTimer( (&hl_pwr_fast), NULL);
	
}

/** @brief Slow Timer Call
 *
 *  Do slow operations for power management:
 *		- Check Bat Voltage
 *		- update status display
 *		- ...
 */
void hl_pwr_slow(void* p){

	/* Start new measurement */
	if( (BatVoltage & HL_PWR_MEASURE_ACTIVE) != HL_PWR_MEASURE_ACTIVE){
		/* restart if no measurement pending, could happen after init */
		hl_adc_init();
		hl_adc_start ();
		BatVoltage |= HL_PWR_MEASURE_ACTIVE;
	}
	if (PowerOffTimer > 0){
		PowerOffTimer--;
	}
}

/** @brief Fast Timer Call
 *
 *  Do fast operations for power management:
 *		- Check bat voltage measurement is ready and finalize
 *		- Manage USB power detection
 *
 */
void hl_pwr_fast(void* p){

	// Check if measurement pending
	if( (BatVoltage & HL_PWR_MEASURE_ACTIVE) == HL_PWR_MEASURE_ACTIVE){
	
		/* read ADC if possible */
		if (hl_adc_isready ()) {
			BatVoltage = hl_adc_convert(hl_adc_result ());
			hl_adc_off();  // save power
		}	
	}


	// disable 5V reg. if Vusb available
	if ( (HL_USB_PWREN_PIN & _BV(HL_USB_PWREN)) == 0){
		// device configured and active
	    // Disable 5V reg.
	    HL_PWR_VCC5_PORT &= ~(_BV(HL_PWR_VCC5));
		hl_pwr_state &= ~HL_PWR_5V_ACTIVE;
		hl_pwr_state |= HL_PWR_USB_ACTIVE;
	}else{
	    // device not configured or in suspend
	    // Enable 5V reg.
	    HL_PWR_VCC5_PORT |= (_BV(HL_PWR_VCC5));		
		hl_pwr_state |= HL_PWR_5V_ACTIVE;
		hl_pwr_state &= ~HL_PWR_USB_ACTIVE;
	}
}

/** @brief 3.3V reulator Control
 *
 *  Switch the 3.3V reg. on or off
 *
 * \param [in] desired state, HL_PWR_ON / HL_PWR_OFF
 */
void hl_pwr_3V3reg(hl_pwr_state_t state){
	if (state == HL_PWR_OFF){
		/* power down 3.3 */
		HL_PWR_VCC33_PORT &= ~(_BV(HL_PWR_VCC33));
		hl_pwr_state &= ~HL_PWR_3V3_ACTIVE;	
	}else{
		/* power up 3.3 */
		HL_PWR_VCC33_PORT |= _BV(HL_PWR_VCC33);	
		hl_pwr_state |= HL_PWR_3V3_ACTIVE;	
	}
}

/** @brief 5V reulator Control
 *
 *  Switch the 5V reg. on or off
 *
 * \param [in] desired state, HL_PWR_ON / HL_PWR_OFF
 */
void hl_pwr_5Vreg(hl_pwr_state_t state){
	if (state == HL_PWR_OFF){
		/* power down 3.3 */
		HL_PWR_VCC5_PORT &= ~(_BV(HL_PWR_VCC5));		
		hl_pwr_state &= ~HL_PWR_5V_ACTIVE;	
	}else{
		/* power up 3.3 */
		HL_PWR_VCC5_PORT |= _BV(HL_PWR_VCC5);	
		hl_pwr_state |= HL_PWR_5V_ACTIVE;	
	}
}

/** @brief LCD HV reulator Control
 *
 *  Switch the LCD High Voltage (~15V) reg. on or off
 *
 * \param [in] desired state, HL_PWR_ON / HL_PWR_OFF
 */
void hl_pwr_LCDreg(hl_pwr_state_t state){
	if (state == HL_PWR_OFF){
		/* power down 3.3 */
		HL_PWR_VLCD_PORT &= ~(_BV(HL_PWR_VLCD));		
		hl_pwr_state &= ~HL_PWR_LCD_ACTIVE;	
	}else{
		/* power up 3.3 */
		HL_PWR_VLCD_PORT |= _BV(HL_PWR_VLCD);	
		hl_pwr_state |= HL_PWR_LCD_ACTIVE;	
	}
}


/** @brief Col0 Key ISR's
 *
 *	Only to generate the wake-up and start the system timer if off
 *
 * 
 *	\todo At the moment only the system timer interrupt is enabled.
 *	It could take up to one second until the sys timer is running.
 *  The start time will depend on the current position of 
 *  the timer2 counter and the Variable FastTimerInc.
 *	We will see later if this is ok, or if it behaves not fluid enough.
 *
 */
ISR(INT0_vect)
{		
	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_KEY);

	/* in case the timer is not running */
	hl_asyn_StartSysTimer();

	/* Now the system is up and running, we do not need any more key interrupts */
	EIMSK &= ~(_BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3));

}

/** @brief Col1 Key ISR's
 *
 *	Only to generate the wake-up and start the system timer if off
 *
 * 
 *	\todo At the moment only the system timer interrupt is enabled.
 *	It could take up to one second until the sys timer is running.
 *  The start time will depend on the current position of 
 *  the timer2 counter and the Variable FastTimerInc.
 *	We will see later if this is ok, or if it behaves not fluid enough.
 *
 */
ISR(INT1_vect)
{		
	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_KEY);

	/* in case the timer is not running */
	hl_asyn_StartSysTimer();

	/* Now the system is up and running, we do not need any more key interrupts */
	EIMSK &= ~(_BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3));


}	

/** @brief Col2 Key ISR's
 *
 *	Only to generate the wake-up and start the system timer if off
 *
 * 
 *	\todo At the moment only the system timer interrupt is enabled.
 *	It could take up to one second until the sys timer is running.
 *  The start time will depend on the current position of 
 *  the timer2 counter and the Variable FastTimerInc.
 *	We will see later if this is ok, or if it behaves not fluid enough.
 *
 */
ISR(INT2_vect)
{		
	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_KEY);

	/* in case the timer is not running */
	hl_asyn_StartSysTimer();

	/* Now the system is up and running, we do not need any more key interrupts */
	EIMSK &= ~(_BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3));
}

/** @brief Col3 Key ISR's
 *
 *	Only to generate the wake-up and start the system timer if off
 *
 *  \note This is the Interrupt is the only interrupt to switch the system on
 *  again when in HL_PWR_ST_OFF
 * 
 *	\todo At the moment only the system timer interrupt is enabled.
 *	It could take up to one second until the sys timer is running.
 *  The start time will depend on the current position of 
 *  the timer2 counter and the Variable FastTimerInc.
 *	We will see later if this is ok, or if it behaves not fluid enough.
 *
 */
ISR(INT3_vect)
{		
	hl_pwr_SetWakeSource(HL_PWR_WAKE_BY_PWR_KEY);

	/* in case the timer is not running */
	hl_asyn_StartSysTimer();

	/* Now the system is up and running, we do not need any more key interrupts */
	EIMSK &= ~(_BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3));
}





/** @brief Try to switch system in the deepest allowed sleep mode
 *
 * 	Function to enter "PwrState".
 *	If a sleep power state is entered, the Control will stay within 
 *  this function until wake-up
 *
 *	\return wake-up source
 *
 *	\note This function must be called outside of an ISR, or special care must be
 *	      taken to ensure the system wake-up, since inside of an ISR the global 
 *        IRQ is disabled.
 *
 */
uint8_t hl_pwr_EnterSleep(void){


	// Check if sleep is allowed






	// Enter sleep mode
	// set_sleep_mode (SLEEP_MODE_IDLE);	
	set_sleep_mode (SLEEP_MODE_PWR_SAVE);  // nur fuer den Moment hard coded
	sleep_enable ();

	// !!! debug	
	PORTB |= 0x04;

	WakeUpSource = HL_PWR_WAKE_NO_SOURCE;
    sleep_cpu ();

	sleep_disable ();

	// Wait for ISR to set wake source
	while (WakeUpSource == HL_PWR_WAKE_NO_SOURCE);

	// !!! debug	
	PORTB &= ~(0x04);

	return WakeUpSource;
}



/** @brief Register a call back function to request power down in higher SW layers
 *
 *	The function will add an entry to the list of functions to be called when the power management
 *  tries to enter a lower power state.
 *  The new entry is added to the front of the list.
 *  
 *  An Application routine can use the call back for the following actions:
 *  - Check if current states or operations require the system to stay on
 *  - set the allowed states
 *
 *  The control should be returned as fast as pssible. Do not wait or complete any pending operations 
 *  within the callback.
 *
 *  @param [in] fkt2call function pointer
 *  @param [in] PwrStAllowed pointer to power state bitmap
 */
 void hl_asyn_RegisterPwrCb(cb_function_t fkt2call, void *PwrStAllowed)
{
  cb_register_call (&cbListPowerFirst, fkt2call, PwrStAllowed);
}
