/* Power Control and Management Functions */
/* Copyright (C) 2009,2010 Thomas Klosa
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file
 * @ingroup geXPal_lib
 * @brief  Power Management System for the eXPal Prototype A
 */


#ifndef POWER_H
#define POWER_H

#include <avr/io.h>

#define HL_PWR_ON	1
#define HL_PWR_OFF	0
#define HL_PWR_FAIL -1
#define HL_PWR_SUCCESS 0

// Bat Voltage Defines
/** @brief Invalid Voltage Value
 *
 *  Voltage unknown or not measured yet
 */ 
#define HL_PWR_VOLT_INVALID		0xffff	// invalid Voltage value

/** @brief MSB in voltage value used to indicate measurement in progress
*/
#define HL_PWR_MEASURE_ACTIVE	0x8000  // Flag for measurement in progress

/*
	Power Information var and bits
*/
typedef int8_t	hl_pwr_state_t;

#define HL_PWR_USB_ACTIVE	0x01	//!< Power from USB is available
#define HL_PWR_3V3_ACTIVE	0x02	//!< 3.3V is active
#define HL_PWR_5V_ACTIVE	0x04	//!< 5V regulator is active
#define HL_PWR_LCD_ACTIVE	0x08	//!< LCD High Voltage is active
#define HL_PWR_BAT_ALCALINE	0x10	//!< Battery type is Alcaline (else 1.2V Accu)

/** @brief PowerState 1  Running at full speed, no sleep at all allowed
 *
 *	The system will operate without sleep. This power state is default when powered from USB.
 *
 */
#define HL_PWR_ST_RUNNING	0x10

/** @brief PowerState 2  Go into idle mode only
 *
 *  Power State with CPU Idle mode.
 *	- CPU Idle Mode, IOclk running, wake on UART and all other sources
 *	- 3.3V Supply state will stay as is
 *	- Display HV on
 *	- not entered when running on USB
 *	- RS232 driver state will stay as is
 *	- 19.53ms timer will stay on
 *
 */
#define HL_PWR_ST_IDLE		0x20

/** @brief PowerState 3  Go into power save mode
 *
 *  Power State with CPU in power save mode.
 *	- CPU Power-Save Mode, IOclk off
 *	- wake on Timer2 and all Key Interrupts
 *	- 3.3V Supply off
 *    - In this mode the CPU can not wake up from USART (no GPS operation possible)
 *    - System must stay in idle mode to access the SD Card
 *	- Display HV on
 *	- not entered when running on USB
 *	- RS232 driver off
 *	- 19.53ms timer off
 *	- 1sec timer on
 *
 *
 */
#define HL_PWR_ST_SLEEP	0x40


/** @brief PowerState 3  Go into sleep mode
 *
 *  Soft Power off state
 *	- CPU Power-Save Mode, IOclk off
 *	- wake on Timer2 and PowerKey Interrupts
 *	- 3.3V Supply off
 *	- Display HV off
 *	- not entered when running on USB
 *	- RS232 driver off
 *	- 19.53ms timer off
 *	- 1sec timer on
 */
#define HL_PWR_ST_OFF	0x80

/*
	The following bits are used with hl_pwr_CtrlAutoSave
	When set, the pwr Management can enter the related power state
*/
typedef int8_t	pwr_info_t;
#define HL_PWR_ALLOW_DOWN	0x20	//!< Power Down allowed
#define HL_PWR_ALLOW_SLEEP	0x40	//!< Sleep Allowed
#define HL_PWR_ALLOW_OFF	0x80	//!< Power Off allowed

#define HL_PWR_OFF_TIME		30		//!< inactivity seconds until power off

/*
	Definition of WakeUp sources
*/
#define HL_PWR_WAKE_BY_KEY		0x01	//!< Key in COL0..2
#define HL_PWR_WAKE_BY_PWR_KEY	0x02	//!< Key in COL3
#define HL_PWR_WAKE_BY_SYS_TMR	0x04	//!< Timer2 COMPA
#define HL_PWR_WAKE_BY_1S_TMR	0x08	//!< Timer2 OV
#define HL_PWR_WAKE_BY_EX232	0x10	//!< USART 3
#define HL_PWR_WAKE_BY_MOD232	0x20	//!< USART 2
#define HL_PWR_WAKE_NO_SOURCE	0x00    //!< Unknown Source 


/** @brief Bitmap variable to indicate the wake-up source
 *
 *	WakeUpSource is a symmary of system actions that caused the system to wake up.
 *	It's cleared before entering sleep, any wake-up ISR will set the individual bit.
 *	When checked some time after wake-up, more than one source might be active
 */
extern volatile uint8_t WakeUpSource;

/** @brief Second timer until power off
 *
 *  If power off is allowed but the timer is not NULL, 
 *  the system will only enter sleep mode.
 */
extern volatile uint16_t PowerOffTimer;

extern volatile hl_pwr_state_t hl_pwr_state;	//!< System power state (current setting)

extern volatile hl_pwr_state_t hl_pwr_info;		//!< Power States,the system is allowed to use

extern volatile uint16_t	BatVoltage;	//!< voltage format: n.f	(n << 8 | (uint8_t)(f*256.0))



/*
	Functions for external use
*/

/** @brief Primary Hardware Init Function
 *
 *	This function has to be called as the first action in main()
 *  It will ensure the correct setting of all processor IO lines and other board hardware.
 *	The function will only set everything in a defined, power saving state and not init any hardware unit to work.
 */
void hl_pwr_InitBoard(void);

/** @brief Init power management system
 *
 * Initialize
 * - Variables
 * - Bat Voltage measurement
 * - Key interrupts
 */
void hl_pwr_InitPwrMan(void);

/** @brief Try to switch system in the deepest allowed sleep mode
 *
 * 	Function to enter "PwrState".
 *	If a sleep power state is entered, the Control will stay within 
 *  this function until wake-up
 *
 *	\return wake-up source
 *
 *	\note This function must be called outside of an ISR, or special care must be
 *	      taken to ensure the system wake-up, since inside of an ISR the global 
 *        IRQ is disabled.
 *
 */
 uint8_t hl_pwr_EnterSleep(void);


/** @brief To be called in ISR to set wake-up source
 *
 * In some cases it might be desired to go back to sleep when the wake-up
 * was caused by a certain source. <BR>
 * Example: Most of the background activities are done within the 20ms timer call.
 *          When the control is returned back to the main loop it might be desired
 *          to call hl_pwr_EnterSleep immediately.
 *
 * @param [in] ws Wake-up source define
 *  HL_PWR_WAKE_BY_KEY <BR>
 *  HL_PWR_WAKE_BY_PWR_KEY <BR>
 *  HL_PWR_WAKE_BY_SYS_TMR <BR>
 *  HL_PWR_WAKE_BY_1S_TMR <BR>
 *  HL_PWR_WAKE_BY_EX232 <BR>
 *  HL_PWR_WAKE_BY_MOD232 <BR>
 *  HL_PWR_WAKE_NO_SOURCE  <BR>
 */
inline void hl_pwr_SetWakeSource(uint8_t ws){

	WakeUpSource |= ws;
}

/** @brief return Wake-up source
*/
inline uint8_t hl_pwr_GetWakeSource(void){
	return WakeUpSource;
}

/** @brief Keep System On
 *
 *	The App has to call this function to avoid automatic power off
 *	otherwise the system will enter HL_PWR_ST_OFF when allowed and the timer
 *	has reached Zero
 *
 */
inline void hl_pwr_AppBusy(void){
	PowerOffTimer = HL_PWR_OFF_TIME;
}

/** @brief Set off timer to zero
 *
 *	The App has to call this function to set the off timer to Zero.
 *	The system will enter HL_PWR_ST_OFF when allowed.
 *
 */
inline void hl_pwr_OffNow(void){
	PowerOffTimer = 0;
}


void hl_pwr_fast(void*); //!< fast system timer call (20ms)
void hl_pwr_slow(void*);  //!< slow system timer call (1sec)

void hl_pwr_3V3reg(hl_pwr_state_t state); //!< Switch 3.3V regulator
void hl_pwr_5Vreg(hl_pwr_state_t state);  //!< Switch 5V regulator
void hl_pwr_LCDreg(hl_pwr_state_t state); //!< Switch LCD HV regulator


/** @brief Request current bat voltage

 @return Voltage to application
         voltage format: n.f	(n << 8 | (uint8_t)(f*256.0))  

*/
inline uint16_t hl_pwr_GetBatVoltage(void){ 
	return (BatVoltage & ~HL_PWR_MEASURE_ACTIVE);
}

/** @brief Return complete power state to application
 * @return Pwr State Bitmap
 *
 */
inline hl_pwr_state_t hl_pwr_GetState(void){ 
	return hl_pwr_state;
}

/** @brief set pwr auto save state from application
 *  @note function will be replaced by a callback
 */
inline void hl_pwr_CtrlAutoSave(pwr_info_t PwrState){

	hl_pwr_info &= ~(HL_PWR_ALLOW_DOWN | HL_PWR_ALLOW_SLEEP | HL_PWR_ALLOW_OFF);
	hl_pwr_info |= (PwrState & (HL_PWR_ALLOW_DOWN | HL_PWR_ALLOW_SLEEP | HL_PWR_ALLOW_OFF));
}




#endif
