/* kbd driver via I2C bus source file */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
/**
 * @file
 * @ingroup gFirstTry
 */

#include <util/twi.h>
#include <util/atomic.h>
#include <avr/interrupt.h>


#include "config.h"
#include "kbd-i2c.h"
#include "events.h"

static inline void
i2c_enable_timer (uint8_t enable)
{
  /* CTC mode */
  TCCR0A = TCCR_BASEVAL | (enable ? TCCR_CLKSRC : 0);
}

static uint8_t kbd_readpending;

static volatile uint8_t *i2c_msgptr;
/* the following is used only in INT context or during init */
static uint8_t i2c_msglen;
static uint8_t i2c_isaddr;
static uint8_t i2c_read;

static uint8_t i2c_acking;

static uint8_t kbd_prevbytes[2];

static const uint8_t kbd_init_cmd[] =
  { 4, KBD_I2C_ADDRESS, 6 /* config */, 0xff, 0xff,
    4, KBD_I2C_ADDRESS, 4 /* polarity */, 0, 0,
    0 /* STOP */, I2C_END_INIT
  };

static uint8_t kbd_read_cmd[] =
  { 2, KBD_I2C_ADDRESS, 0 /* input */,
    3, KBD_I2C_ADDRESS|1, /* placeholder */0,0,
    0 /* STOP */, I2C_END_KBD
  };

/* frame format
 * reading:
 *  START
 *   0 addr|0 (address byte, write)
 *   1 cmd    (READ command, select register)
 *  RESTART
 *   2 addr|1 (address byte, read)
 *   3 byte0
 *   4 byte1
 *  STOP

 * writing:
 *  START
 *   0 addr|0 (address byte, write)
 *   1 cmd    (WRITE command, select register)
 *   2..      register contents to write
 *  STOP
 */

static void
i2c_start_transfer (i2c_msg_t *m, uint8_t in_interrupt)
{
  if (in_interrupt)
    NONATOMIC_BLOCK (NONATOMIC_FORCEOFF) {
      while (1) { /* wait until prev msg complete */
	uint8_t msg_sent;
	ATOMIC_BLOCK (ATOMIC_FORCEON) {
	  msg_sent = i2c_msgptr == 0;
	}
	if (msg_sent) break;
      }
    }
  else
    while (1) { /* wait until prev msg complete */
      uint8_t msg_sent;
      ATOMIC_BLOCK (ATOMIC_FORCEON) {
	msg_sent = i2c_msgptr == 0;
      }
      if (msg_sent) break;
    }

  i2c_acking = 0;
  /* init ptrs */
  i2c_msglen = m->msglen;
  i2c_msgptr = &(m->msg[0]);
  i2c_isaddr = 1;
  i2c_read = m->msg[0] & 1;
  /* first byte is always a write */
  KBD_I2C_PORT_D |= _BV(KBD_I2C_SDA_D);
  
  /* Reset and start timer */
  TCNT0 = 0;
  TIFR0 |= OCF0A;
  i2c_enable_timer (1);
}

uint8_t
kbd_i2c_init (void)
{
  /* configure SCL and SDA as open collector */
  KBD_I2C_PORT_D |= _BV(KBD_I2C_SCL_D) | _BV(KBD_I2C_SDA_D);
  /* set TWI operation mode for USI,
   * clock source: software */
  USICR = _BV(USIOIE) | _BV(USIWM1);

  /* we need one timer to toggle the SCL line
   * for toggling frequency 100kHz we need to toggle every 80 clocks */
  OCR0A = F_CPU/F_TWI/2;
  i2c_enable_timer (0);
  /* we only enable timer INT during TWI ops (got key INT, setup of chip) */
  TIMSK0 = 0;

  i2c_start_transfer ((i2c_msg_t*)kbd_init_cmd,0);
  return 1;
}

static void
kbd_put_i2c_complete_event (uint8_t byte0, uint8_t byte1)
{
  uint8_t newstate[2] = {byte0,byte1};
  uint8_t changed [2];
  uint8_t i;
  uint16_t c,k;
  for (i=0; i<2; i++) {
    changed[i] = newstate[i] ^ kbd_prevbytes[i];
    kbd_prevbytes[i] = newstate[i];
  }
  c = changed[0] | (changed[1]<<8);
  k = newstate[0] | (newstate[1]<<8);
  k = ~k; /* key signals are active LOW */
  put_key_event (c, k);
}

static void
kbd_i2c_complete_init (void)
{
  /* enable Pin Change Interrupt 6 */
  EICRA = 0; /* INT on low level */
  PCMSK0 |= PCINT6;
  EIMSK |= PCIE0;
}

static void
i2c_complete (void)
{
  switch (*i2c_msgptr) {
  case I2C_END_KBD:
    kbd_put_i2c_complete_event (i2c_msgptr[-3],i2c_msgptr[-2]);
    break;
  case I2C_END_INIT:
    kbd_i2c_complete_init ();
    break;
  }
  i2c_msgptr = 0;
}

ISR(USI_OVERFLOW_vect)
{
  /* TODO: Generate START/STOP condition */
  i2c_enable_timer (0);
  if (i2c_acking) {
    i2c_acking = 0;
    /* prepare next byte */
    if (--i2c_msglen == 0) {
      i2c_msglen = *(i2c_msgptr++);
      if (i2c_msglen == 0) {
	i2c_complete ();
	return; /* don't start timer */
      }
      i2c_isaddr = 1;
      i2c_read = i2c_msgptr[0] & 1;
    }
    if (i2c_isaddr || !i2c_read) {
      /* put byte into shift register */
      USIDR = *(i2c_msgptr++);
      /* prepare for sending byte */
      KBD_I2C_PORT_D |= _BV(KBD_I2C_SDA_D);
    } else {
      /* prepare for reading byte */
      KBD_I2C_PORT_D &= ~_BV(KBD_I2C_SDA_D);
    }
  } else {
    i2c_acking = 1;

    if (i2c_isaddr || !i2c_read) {
      /* prepare for reading ACK */
      KBD_I2C_PORT_D &= ~_BV(KBD_I2C_SDA_D);
    } else {
      /* get byte */
      *(i2c_msgptr++) = USIDR;
      /* prepare for sending ACK */
      KBD_I2C_PORT_D |= _BV(KBD_I2C_SDA_D);
    }
    /* prepare ACK */
    USISR = _BV(USIOIF) | _BV(USICNT3) | _BV(USICNT2) | _BV(USICNT1);
    USIDR = 0x0;
  }
  i2c_enable_timer (1);
}

ISR(TIMER0_OVF_vect)
{
  uint8_t before = KBD_I2C_PORT;
  USICR |= _BV(USITC); /* toggle pin */
  /* when SCL is to go high, wait for it */
  if ((before & _BV(KBD_I2C_SCL)) == 0) {
    i2c_enable_timer (0);
    /* FIXME: busy-wait for indefinite time in INT routine */
    while ( (KBD_I2C_PORT & _BV(KBD_I2C_SCL)) == 0);
    i2c_enable_timer (1);
  }
}

ISR(PCINT0_vect)
{
  if (KBD_I2C_PORT & _BV(KBD_I2C_INT)) {
      i2c_start_transfer ((i2c_msg_t*)kbd_read_cmd, 1);
  }
}
