// This file has been prepared for Doxygen automatic documentation generation.
/*! \file *********************************************************************
 *
 * \brief  Single UART driver source file
 *
 *      The UART driver provides an interface to one of the UART peripherals.
 *      The default implementation for the DB101 firmware is to interface the
 *      UART1 module, which is connected to the external UART connections of the
 *      DB101 board.
 *
 *      The driver provides the basic functionality for interrupt controlled
 *      reception and transmission through FIFO queues from the FIFO Queue
 *      Library, and of course to configure the UART baud rate, parity mode etc.
 *
 *      The driver source code is designed to support only one UART module,
 *      which means that it cannot change UART modules runtime. This limitation
 *      makes the driver very speed efficient, but less flexible. For the DB101
 *      board, this is not an issue, since the other UART module is solely used
 *      for accessing the on-board DataFlash memory device.
 *
 * \par Application note:
 *      AVR482: DB101 Software
 *
 * \par Documentation
 *      For comprehensive code documentation, supported compilers, compiler
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com
 *
 * $Revision: 2412 $
 * $URL: http://revisor.norway.atmel.com/AppsAVR8/avr482_db101_software_users_guide/tags/release_A_code_20070917/IAR/uart_driver/uart_driver.h $
 * $Date: 2007-09-17 01:36:31 -0600 (ma, 17 sep 2007) $  \n
 *
 * Copyright (c) 2006, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#ifndef UART_DRIVER_H
#define UART_DRIVER_H

#include <cal.h>
#include <common.h>
#include <stdint.h>
#include "fifo_lib.h"


/*********************
 * Defines and types.
 *********************/


#define UART_CALC_UBRR_VALUE(baud) ((CPU_F / 16 / (baud)) - 1)


/* Global vars (access through macros) */
typedef struct flags_struct{
	unsigned int rxOverflow : 1;
} UART_flags_t;
extern UART_flags_t volatile UART_flags;


typedef enum UART_parityMode_enum {
	UART_PARITY_NONE,
	UART_PARITY_EVEN,
	UART_PARITY_ODD
} UART_parityMode_t;


/***********************
 * Function prototypes.
 ***********************/

//! Initialize a UART, by setting baud rate and enabling Rx and Tx.
void UART_Init( FIFO_handle_t * rxFifoHandle, FIFO_handle_t * txFifoHandle );

//! Set Baud rate for the UART
void UART_SetBaudRateValue( uint16_t ubrrValue );
//! Set parity mode for the UART
void UART_SetParityMode( UART_parityMode_t parityMode );
//! Set number of data bits
void UART_SetDataBits( uint8_t dataBitCount );
//! Set number of stop bits
void UART_SetStopBits( uint8_t stopBitCount );

//! Set the RX fifo to be used by the UART.
void UART_SetRxFifo(FIFO_handle_t * rxFifoHandle);
//! Set the TX fifo to be used by the UART.
void UART_SetTxFifo(FIFO_handle_t * txFifoHandle);
//! Get the RX fifo currently used by the UART.
FIFO_handle_t * UART_GetRxFifo( void );
//! Get the TX fifo currently used by the UART.
FIFO_handle_t * UART_GetTxFifo( void );



/********************************************
 * Convenience macros - used function style.
 ********************************************/

#define UART_StartTransmission() UART_EnableUdreInterrupt()     //!< If data are avaialble in the Tx buffer, a transmissions is started (enables the UDRE interrupt).
#define UART_StopTransmission() UART_DisableUdreInterrupt()     //!< Disables the UDRE interrupt to stop an ongoing transmissions. Can be used when adding a datapackage to the TX fifo that should be transmitted in full, and not partially.

#define UART_SetRxOverflowFlag() UART_flags.rxOverflow = 1
#define UART_GetRxOverflowFlag() (UART_flags.rxOverflow == 1)

#define UART_EnableRxInterrupt() UART_CTRLB |= (1<<RXCIE)       //!< Enable RXC interrupt.
#define UART_EnableTxInterrupt() UART_CTRLB |= (1<<TXCIE)       //!< Enable TXC interrupt.
#define UART_EnableUdreInterrupt() UART_CTRLB |= (1<<UDRIE)     //!< Enable UDRE interrupt.
#define UART_DisableRxInterrupt() UART_CTRLB &= ~(1<<RXCIE)     //!< Disable RX interrupt.
#define UART_DisableTxInterrupt() UART_CTRLB &= ~(1<<TXCIE)     //!< Disable TXC interrupt.
#define UART_DisableUdreInterrupt() UART_CTRLB &= ~(1<<UDRIE)   //!< Disable UDRE interrupt.

#endif
