/* ADC functions */
/* Copyright (C) 2009,2010 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef HLOG_ADC_H
#define HLOG_ADC_H

/*!\file
 * \ingroup geXPal_lib
 * \brief access to ADC and analog multiplexer
 */

#include <stdint.h>

/*! \brief decide which channel to sample
 *
 * available channels:
 */
typedef enum {
  /*! ADC0, connected to battery via 1 MOhm/1 MOhm voltage divider
   */
  adc_chan0,
  /*! 1.1 V internal reference voltage */
  adc_ref,
} hl_adc_channel;

/*! \brief prepare ADC for use
 *
 * hl_adc_init() powers up the ADC,
 * defines PF0 as analog input,
 * sets the ADC prescaler and selects left-adjustment for the result.
 */
void hl_adc_init (void);

/*! \brief switch off ADC
 *
 * hl_adc_off() powers down the ADC to conserve power between sampling.
 */
void hl_adc_off (void);

/*! \brief select channel to sample
 */
void hl_adc_select (hl_adc_channel chan);
/*! \brief start ADC sampling process
 */
void hl_adc_start (void);
/*! \brief check whether sampling process has finished
 */
uint8_t hl_adc_isready (void);
/*! \brief get result from ADC
 */
uint16_t hl_adc_result (void);

/*!
 * \brief convert ADC result according to builtin factors
 *
 * \param r output from ADC
 * \return voltage, integer part in high byte,
 *         decimal fractional part in low byte
 *
 * You can print the result like this:
 *
 * v = hl_adc_convert (r);
 * printf ("%d.%02d", v >> 8, v & 0xFF);
 */
inline uint16_t hl_adc_convert (uint16_t r)
{
  /* reference voltage 2.56 V */
  /* voltage divider 1M/1M (factor 0.5) */
  uint32_t r2 = r;
  r2 *= 512;
  return (r2/100)>>8;
}

#endif
