/* LCD graphics functions */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef DISPLAY_H
#define DISPLAY_H

#include <stdint.h>
#include <avr/pgmspace.h>
#include "../../font/font.h"

/*!\file
 * \ingroup geXPal_lib
 * \brief mid-level display functions
 *
 * This file declares functions for display access: simple graphics
 * and text display.
 *
 * More complex graphics functions like arbitrary lines, circles
 * or ellipses will be implemented when we deem them necessary.
 */

/*!\brief ROPs for one pixel: */
typedef enum {
  op_clear  = 0b00, /*!<\brief clear pixel */
  op_invert = 0b01, /*!<\brief invert pixel */
  op_keep   = 0b10, /*!<\brief keep pixel as-is */
  op_set    = 0b11, /*!<\brief set pixel */
} hl_blt_1op;

/* for two pixels: */
typedef enum {
  op2_clearall = 0b0000,
  op2_notand   = 0b0001,
  op2_0010     = 0b0010,
  op2_notsrc   = 0b0011,
  op2_clearnot = 0b0100,
  op2_invert   = 0b0101,
  op2_xor      = 0b0110,
  op2_notor    = 0b0111,
  op2_and      = 0b1000,
  op2_1001     = 0b1001,
  op2_keep     = 0b1010,
  op2_1011     = 0b1011,
  op2_copysrc  = 0b1100,
  op2_1101     = 0b1101,
  op2_or       = 0b1110,
  op2_setall   = 0b1111,
} hl_blt_2op;

/*!\brief operate on a single pixel
 *
 * hl_lcd_op_pixel() operates on a single pixel
 * given by its coordinates
 * \param x X coordinate of pixel
 * \param y Y coordinate of pixel
 * \param op what to do to given pixel
 */
void hl_lcd_op_pixel (uint8_t x, uint8_t y, hl_blt_1op op);

/*!\brief fill all display bytes with given pattern
 *
 * \param fill 8-bit byte pattern to write to display
 */
void hl_lcd_setall (uint8_t fill);

/*!\brief select font to use for text operations
 */
void hl_lcd_set_font (const font_t * f);

/*!\brief draw a character at pixel position (x,y)
 *
 * hl_lcd_drawchar() draws a character at an arbitrary position
 * \param x X coordinate of upper left character box corner
 * \param y Y coordinate of upper left character box corner
 * \param c character to draw (pattern and size is defined by previous call
 *          to hl_lcd_set_font())
 */
void hl_lcd_drawchar (uint8_t x, uint8_t y, char c);

/*!\brief get width (in pixels) of currently selected font
 */
uint8_t hl_lcd_get_charwidth (void) __attribute__((pure));
/*!\brief get height (in pixels) of currently selected font
 */
uint8_t hl_lcd_get_charheight (void) __attribute__((pure));

/*!\brief draw a character string beginning at pixel position (x,y)
 *
 * hl_lcd_drawstr() may be implemented in a more efficient way
 * than calling hl_lcd_drawchar multiple times, especially for
 * char_width != 8 pixel  or  x%8 != 0
 */
void hl_lcd_drawstr (uint8_t x, uint8_t y, const char * s);
/*!\brief like hl_lcd_drawstr(), but for strings stored in Flash memory
 */
void hl_lcd_drawstr_P (uint8_t x, uint8_t y, PGM_P s);

/*!\brief clear complete LCD display
 */
inline void hl_lcd_clear (void) { hl_lcd_setall (0); }

/*!\brief scroll display up
 *
 * hl_lcd_scroll_up() scrolls up the display for the specified number
 * of lines
 * \param lines lines to scroll up (in pixels)
 */
void hl_lcd_scroll_up (uint8_t lines);

/*!\brief fill/clear/invert/keep all pixels inside a rectangle
 *
 * (x0,y0) is inclusive, (x1,y1) is exclusive
 */
void hl_lcd_set_rectangle (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
			   hl_blt_1op op);

/*!\brief fill/clear/invert/keep a single horizontal display line
 *
 * This is a special case of hl_lcd_set_rectangle().
 */
inline void hl_lcd_set_line_horiz (uint8_t x0, uint8_t y, uint8_t x1,
				   hl_blt_1op op)
{
  hl_lcd_set_rectangle (x0,y, x1,y+1, op);
}

/*!\brief fill/clear/invert/keep a single horizontal display column
 *
 * This is a special case of hl_lcd_set_rectangle().
 */
inline void hl_lcd_set_line_vert (uint8_t x, uint8_t y0, uint8_t y1,
				  hl_blt_1op op)
{
  hl_lcd_set_rectangle (x,y0, x+1,y1, op);
}


/*!\brief save display area for later re-display at same position
 *
 * DO NOT RELY ON THE FORMAT OF THE STORED DATA!
 * This may change depending on implementation and even
 * depending on the position paramaters. You may copy the data
 * around or discard it. Restore is only guaranteed to work to
 * the exact same position and size the data was saved from.
 */
size_t hl_lcd_save (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
		    uint8_t * save, size_t size);
/*!\brief restore display area stored by hl_lcd_save() at same position
 */
size_t hl_lcd_restore (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
		       const uint8_t * save, size_t size);
/*!\brief calculate necessary size of buffer for hl_lcd_save()*/
size_t hl_lcd_save_size (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1)
  __attribute__((const));
/*!\brief same as hl_lcd_restore(), but as macro */
#define HL_LCD_SAVE_SIZE(X0,Y0, X1,Y1) \
  ((((X1+7)/8)-(X0/8))*(Y1-Y0))


#endif
