/* FIFO declarations */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef FIFO_H
#define FIFO_H

#include <avr/interrupt.h>

typedef uint8_t fifo_item_t;
typedef uint8_t fifo_size_t;

typedef struct fifo_s {
  fifo_item_t * buffer;
  fifo_size_t size;
  fifo_size_t head;
  fifo_size_t tail;
  fifo_size_t num_items;
  uint8_t overflow:1;
  uint8_t underflow:1;
} fifo_t;

void fifo_init (fifo_t * f, fifo_item_t * buffer, fifo_size_t size);
inline fifo_item_t * fifo_get_buffer (fifo_t * f) { return f->buffer; }

inline uint8_t fifo_is_empty (const fifo_t * f) { return f->num_items == 0; }
inline uint8_t fifo_is_full (const fifo_t * f) { return f->num_items == f->size; }

/* Assume interrupts are turned off */
inline void fifo_putQ_item (fifo_t * f, fifo_item_t c)
{
  fifo_size_t tail, head;
  tail = f->tail;
  head = f->head;
  if (f->num_items < f->size) {
    /* still room in buf */
    f->buffer[head++] = c;
    if (head == f->size) {
      head = 0;
    }
    f->head = head;
    f->num_items++;
  } else {
    /* set overflow flag, char is lost */
    f->overflow = 1;
  }
}

inline fifo_item_t fifo_getQ_item (fifo_t * f)
{
  fifo_size_t tail;
  tail = f->tail;
  if (f->num_items) {
    fifo_item_t data = f->buffer[tail++];
    if (tail == f->size) {
      tail = 0;
    }
    f->tail = tail;
    f->num_items--;
    return data;
  }
  f->underflow = 1;
  return 0;
}

fifo_item_t fifo_get_item (fifo_t * f);
inline uint8_t fifo_get_uint8 (fifo_t * f)
{ return (uint8_t) fifo_get_item (f); }

uint8_t fifo_put_item (fifo_t * f, fifo_item_t i);
inline uint8_t fifo_put_uint8 (fifo_t * f, uint8_t b)
{ return fifo_put_item (f, (fifo_item_t)b); }

#endif
