/* usart interface header file */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef USARTS_H
#define USARTS_H

#include <stdint.h>
#include <avr/pgmspace.h>
#include <avr/io.h>

#include "fifo.h"

/*!\file
 * \ingroup geXPal_lib
 * \brief USART access
 */

/*! \brief control and status for one USART */
typedef struct usart_control_s {
  /*!\brief address of first USART register */
  volatile uint8_t * const base;
  /*!\brief send buffer */
  fifo_t txbuf;
  /*!\brief receive buffer */
  fifo_t rxbuf;
  /*!\brief address of power reduction register responsible for this USART */
  volatile uint8_t * const prr;
  /*!\brief bitmask in the power register responsible for this USART */
  uint8_t const prusart; /* bit mask */
  /*!\brief Name for this USART (string stored in FLASH) */
  prog_char /*PROGMEM*/ * const name;
} usart_control_t;

/*!\brief selector for USART port */
typedef enum {
  usart_ext, /*!<\brief external port, uses RS-232 level converter */
  usart_mod, /*!<\brief GPS module mounted on board */
} hlog_usart_t;

/* \brief initialize USART
 *
 * initializes one usart_control_s structure and powers up the
 * USART and, if applicable, the level converter.
 * \param which selects which USART to use
 * \param txbuf pointer to a FIFO buffer used for sending
 * \param txsize size of txbuf
 * \param rxbuf pointer to a FIFO buffer used for receiving
 * \param rxsize size of rxbuf
 * \return pointer to the usart_control_t structure used for the selected
 *         USART
 *
 * If txbuf is NULL, the USART will not be available for sending.
 * If rxbuf is NULL, the USART will not be available for receiving.
 * 
 * There will be one and only one usart_control_s structure for
 * each usable USART.
 */
usart_control_t * usart_init (hlog_usart_t which,
			      fifo_item_t * txbuf, fifo_size_t txsize,
			      fifo_item_t * rxbuf, fifo_size_t rxsize);

/*! \brief set baud rate for USART
 *
 * \param baud wanted rate
 * \return actual rate, 0 if not possible
 */
uint16_t usart_set_baudrate (usart_control_t * uc, uint16_t baud);

/*!\brief disable USART */
void usart_disable (usart_control_t *);
/*!\brief power-off USART
 *
 * This puts the USART into power-down mode
 */
void usart_poweroff (usart_control_t *);

/*!\brief enable USART
 *
 * This actually switches on the USART indicated by uc.
 * \param uc
 */
void usart_enable (usart_control_t * uc);

/*!\brief get one byte from USART (blocking) */
uint8_t usart_recv (usart_control_t * uc);

/*!\brief check if usart_recv() will not block */
inline uint8_t usart_can_recv (usart_control_t * uc)
{
  return !fifo_is_empty(&uc->rxbuf);
}

/*!\brief send one byte to USART */
void usart_send (usart_control_t *, uint8_t);

/*! \brief check whether external RS232 interface is connected
 */
uint8_t usart_ext_is_connected (void);

/*!\brief check if usart_send() will not block */
inline uint8_t usart_can_send (usart_control_t * uc)
{
  return !fifo_is_full(&uc->rxbuf);
}

#endif
