/* font storage and access */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <avr/pgmspace.h>

#include "display.h"
#include "lh155.h"

#include "../../font/font.h"

/* make a copy in RAM for faster, simpler access */
static font_t current_font;
static struct {
  uint8_t char_width_bytes; /* number of bytes in one horizontal line */
  uint8_t char_bytes; /* number of bytes per character */
} current_priv;

void hl_lcd_set_font (const font_t * f)
{
  font_t * cf = &current_font;
  memcpy_P (cf, f, sizeof (current_font));
  current_priv.char_width_bytes = (cf->char_width+(uint8_t)7)/(uint8_t)8;
  current_priv.char_bytes = current_priv.char_width_bytes * cf->char_height;
}

uint8_t hl_lcd_get_charwidth (void)
{
  return current_font.char_width;
}

uint8_t hl_lcd_get_charheight (void)
{
  return current_font.char_height;
}

#undef DUMMY1READ

/* draw char at pixel position (x,y)
 * NOTE: at the moment, only positions with x%CHAR_WIDTH == 0 are supported
 */
void hl_lcd_drawchar (uint8_t x, uint8_t y, char c)
{
  const font_t * cf = &current_font;
  const char *p = &(cf->font_data[(uint8_t)c * current_priv.char_bytes]);
  const char *p1;
  uint8_t bit, mask, end_bit;
  uint8_t width = cf->char_width;
  uint8_t height = cf->char_height;
  int8_t i;
  uint8_t end_x = x + width;
  uint8_t end_y = y + height;

  HL_LCD_WriteCmd (HL_LCD_INC_CTRL | HL_LCD_INC_CTRL_AIM | HL_LCD_INC_CTRL_AYI);

  bit = hl_pixel_setXY_bit (x, y);
  mask = - (1 << bit);
  p1 = p;

  x /= 8;
  if (bit + width <= 8) {
    /* restricted to a single byte column */
    if (bit + width < 8)
      mask &= ~ (- (1 << (bit+width)));

    if (~mask) {
#ifdef DUMMY1READ
      HL_LCD_ReadData (); /* dummy read */
#endif

      for (i=0; i<height; i++) {
	uint8_t new = mask & (pgm_read_byte (p1) << bit);
#ifndef DUMMY1READ
	HL_LCD_ReadData (); /* dummy read */
#endif
	uint8_t tmp = HL_LCD_ReadData ();
	tmp &= ~mask;
	HL_LCD_WriteData (tmp | new);
	p1 += 1 /*current_priv.char_width_bytes*/;
      }
    } else {
      /* we can just copy the bytes into the display RAM */
      for (i=0; i<height; i++) {
	uint8_t new = (pgm_read_byte (p1));
	HL_LCD_WriteData (new);
	p1 += 1 /*current_priv.char_width_bytes*/;
      }
    }
  } else {
    /* multi-column */

    end_bit = end_x % 8;

    end_x /= 8;

    if (~mask) {
#ifdef DUMMY1READ
      HL_LCD_ReadData (); /* dummy read */
#endif
      /* leftmost part ( 000aaaaa ) */
      for (i=0; i<height; i++) {
	uint8_t new = mask & (pgm_read_byte (p1) << bit);
#ifndef DUMMY1READ
	HL_LCD_ReadData (); /* dummy read */
#endif
	uint8_t tmp = HL_LCD_ReadData ();
	tmp &= ~mask;
	HL_LCD_WriteData (tmp | new);
	p1 += current_priv.char_width_bytes;
      }
    } else {
      /* we can just copy the bytes into the display RAM */
      for (i=0; i<height; i++) {
	uint8_t new = (pgm_read_byte (p1));
	HL_LCD_WriteData (new);
	p1 += current_priv.char_width_bytes;
      }
    }

    x++;
    /* middle part ( aaabbbbb ) */
    for (;x < end_x; x++, p++) {
      /* restart at top */
      HL_LCD_SetX (x);
      HL_LCD_SetY (y);

      if (~mask) {
	for (i=0, p1 = p; i<height; i++) {
	  uint8_t new = ~mask & (pgm_read_byte (p1) >> (8-bit));
	  new |= mask & (pgm_read_byte (p1+1) << bit);
	  HL_LCD_WriteData (new);

	  p1 += current_priv.char_width_bytes;
	}
      } else {
	for (i=0, p1 = p; i<height; i++) {
	  uint8_t new = (pgm_read_byte (p1+1));
	  HL_LCD_WriteData (new);

	  p1 += current_priv.char_width_bytes;
	}
      }
    }

    
    mask = - (1 << end_bit);
    /* last part ( bbb00000 ) */
    /* restart at top */
    HL_LCD_SetX (x);
    HL_LCD_SetY (y);

    if (bit == 0) {
      bit = 8;
      p++;
    }

    if (mask) {
#ifdef DUMMY1READ
      HL_LCD_ReadData (); /* dummy read */
#endif
      for (i=0, p1 = p; i<height; i++) {
	uint8_t new = ~mask & (pgm_read_byte (p1) >> (8-bit));
#ifndef DUMMY1READ
      HL_LCD_ReadData (); /* dummy read */
#endif
	uint8_t tmp = HL_LCD_ReadData ();
	tmp &= mask;
	HL_LCD_WriteData (tmp | new);
	p1 += current_priv.char_width_bytes;
      }
    } else {
      for (i=0, p1 = p; i<height; i++) {
	uint8_t new = (pgm_read_byte (p1) >> (8-bit));
	HL_LCD_WriteData (new);
	p1 += current_priv.char_width_bytes;
      }
    }
  }
}

void hl_lcd_drawstr (uint8_t x, uint8_t y, const char * s)
{
  char c;
  while ((c = *s++)) {
    hl_lcd_drawchar (x, y, c);
    if (x > HL_LCD_WIDTH-current_font.char_width)
      break; /* don't wrap around */
    x += current_font.char_width;
  }
}

void hl_lcd_drawstr_P (uint8_t x, uint8_t y, PGM_P s)
{
  char c;
  while ((c = pgm_read_byte (s++))) {
    hl_lcd_drawchar (x, y, c);
    if (x > HL_LCD_WIDTH-current_font.char_width)
      break; /* don't wrap around */
    x += current_font.char_width;
  }
}

