/* rectangle operation */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "display.h"
#include "lh155.h"

/* fill/clear/invert/keep all pixels inside a rectangle
 * (x0,y0) is inclusive, (x1,y1) is exclusive
 */
void hl_lcd_set_rectangle (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
			   hl_blt_1op op)
{
  /* TODO: get constant exprs out of inner loops */

  uint8_t x0b, x1b; /* x[01] as byte address */
  uint8_t bit0, bit1; /*     as bit number in corresponding byte */
  uint8_t mask;
  
  if (op == op_keep) return; /* easy - nothing to do */
  
  /* crop rectangle to visible display area */
  if (x1 > HL_LCD_WIDTH) x1 = HL_LCD_WIDTH;
  if (y1 > HL_LCD_HEIGHT) y1 = HL_LCD_HEIGHT;
  
  /* do nothing if start >= end */
  if (x0 >= x1 || y0 >= y1) return;

  x0b = x0/8;
  x1b = (x1-(uint8_t)1)/8; /* x1b points to last byte to be written */

  bit0 = x0%8;
  bit1 = (x1-(uint8_t)1)%8;

  HL_LCD_WriteCmd (HL_LCD_INC_CTRL | HL_LCD_INC_CTRL_AIM
		   | HL_LCD_INC_CTRL_AYI);

  if (x0b == x1b) {
    /* just one column of bytes will be changed */
    HL_LCD_SetX (x0b);
    HL_LCD_SetY (y0);

    /* start */
    mask = -(1<<bit0);
    /* end */
    mask &= ~(-(2<<bit1));

    for (; y0 < y1; y0++) {
      uint8_t r;
      HL_LCD_ReadData (); /* dummy read */
      r = HL_LCD_ReadData ();

      switch (op) {
      case op_clear:  r &= ~mask; break;
      case op_set:    r |= mask; break;
      case op_invert: r ^= mask; break;
      default: /* should not happen */ break;
      }
      HL_LCD_WriteData (r);
    }
  } else {
    /* at least two columns will be changed */
    uint8_t y;

    /* first (incomplete) column */
    if (bit0) {
      /* start */
      mask = -(1<<bit0);

      HL_LCD_SetX (x0b);
      HL_LCD_SetY (y0);

      for (y = y0; y < y1; y++) {
	uint8_t r;
	HL_LCD_ReadData (); /* dummy read */
	r = HL_LCD_ReadData ();
	
	switch (op) {
	case op_clear:  r &= ~mask; break;
	case op_set:    r |= mask; break;
	case op_invert: r ^= mask; break;
	default: /* should not happen */ break;
	}
	HL_LCD_WriteData (r);
      }
      x0b++;
    }

    /* complete column(s) */
    for (; x0b < x1b; x0b++) {
      HL_LCD_SetX (x0b);
      HL_LCD_SetY (y0);
      
      for (y = y0; y < y1; y++) {
	uint8_t r;
	
	switch (op) {
	case op_clear:  r = 0; break;
	case op_set:    r = 0xFF; break;
	case op_invert:
	  HL_LCD_ReadData (); /* dummy read */
	  r = HL_LCD_ReadData ();
	  r = ~r;
	  break;
	default: /* should not happen */
	  r = HL_LCD_ReadData (); /* shut up compiler warning */
	  break;
	}
	HL_LCD_WriteData (r);
      }
    }
    
    /* last column */

    /* end */
    mask = ~(-(2<<bit1));
    
    HL_LCD_SetX (x0b);
    HL_LCD_SetY (y0);

    for (y = y0; y < y1; y++) {
      uint8_t r;
      HL_LCD_ReadData (); /* dummy read */
      r = HL_LCD_ReadData ();
	
      switch (op) {
      case op_clear:  r &= ~mask; break;
      case op_set:    r |= mask; break;
      case op_invert: r ^= mask; break;
      default: /* should not happen */ break;
      }
      HL_LCD_WriteData (r);
    }
  }
  
  HL_LCD_WriteCmd (HL_LCD_INC_CTRL | 0);
}

