/* save/restore display area operation */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "display.h"
#include "lh155.h"

size_t hl_lcd_save_size (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1)
{
  return HL_LCD_SAVE_SIZE(x0,y0, x1,y1);
}


/* save or restore a display area
 * (x0,y0) is inclusive, (x1,y1) is exclusive
 */
static inline
size_t hl_lcd_save_restore (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
			    uint8_t * save,const uint8_t * restore,size_t size)
  __attribute__((always_inline));

static inline
size_t hl_lcd_save_restore (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
			    uint8_t * save,const uint8_t * restore,size_t size)
{
  /* TODO: get constant exprs out of inner loops */

  uint8_t x0b, x1b; /* x[01] as byte address */
  uint8_t bit0, bit1; /*     as bit number in corresponding byte */
  uint8_t mask;
  
  
  /* crop rectangle to visible display area */
  if (x1 > HL_LCD_WIDTH) x1 = HL_LCD_WIDTH;
  if (y1 > HL_LCD_HEIGHT) y1 = HL_LCD_HEIGHT;
  
  /* do nothing if start >= end */
  if (x0 >= x1 || y0 >= y1) return 0;

  /* size checks */
  if (size == 0) {
    return hl_lcd_save_size (x0,y0, x1,y1);
  }
  if (hl_lcd_save_size (x0,y0, x1,y1) > size) {
    return -1;
  }

  x0b = x0/8;
  x1b = (x1-(uint8_t)1)/8; /* x1b points to last byte to be written */

  bit0 = x0%8;
  bit1 = (x1-(uint8_t)1)%8;

  HL_LCD_WriteCmd (HL_LCD_INC_CTRL | (restore ? HL_LCD_INC_CTRL_AIM : 0)
		   | HL_LCD_INC_CTRL_AYI);

  if (x0b == x1b) {
    /* just one column of bytes will be changed */
    HL_LCD_SetX (x0b);
    HL_LCD_SetY (y0);

    /* start */
    mask = -(1<<bit0);
    /* end */
    mask &= ~(-(2<<bit1));

    if (!restore) HL_LCD_ReadData (); /* dummy read */

    for (; y0 < y1; y0++) {
      uint8_t r;
      if (restore) {
	HL_LCD_ReadData (); /* dummy read */
	r = HL_LCD_ReadData ();
	r &= ~mask;
	r |= (*restore++) & mask;
	HL_LCD_WriteData (r);
      } else {
	r = HL_LCD_ReadData ();
	*save++ = r & mask;
      }
    }
  } else {
    /* at least two columns will be used */
    uint8_t y;

    /* first (incomplete) column */
    if (bit0) {
      /* start */
      mask = -(1<<bit0);

      HL_LCD_SetX (x0b);
      HL_LCD_SetY (y0);

      if (!restore) HL_LCD_ReadData (); /* dummy read */

      for (y = y0; y < y1; y++) {
	uint8_t r;
	if (restore) {
	  HL_LCD_ReadData (); /* dummy read */
	  r = HL_LCD_ReadData ();
	  r &= ~mask;
	  r |= (*restore++) & mask;
	  HL_LCD_WriteData (r);
	} else {
	  r = HL_LCD_ReadData ();
	  *save++ = r & mask;
	}
      }
      x0b++;
    }

    /* complete column(s) */
    for (; x0b < x1b; x0b++) {
      HL_LCD_SetX (x0b);
      HL_LCD_SetY (y0);
      
      if (!restore) HL_LCD_ReadData (); /* dummy read */

      for (y = y0; y < y1; y++) {
	uint8_t r;
	if (restore) {
	  r = *restore++;
	  HL_LCD_WriteData (r);
	} else {
	  r = HL_LCD_ReadData ();
	  *save++ = r;
	}
      }
    }
    
    /* last column */

    /* end */
    mask = ~(-(2<<bit1));
    
    HL_LCD_SetX (x0b);
    HL_LCD_SetY (y0);

    if (!restore) HL_LCD_ReadData (); /* dummy read */

    for (y = y0; y < y1; y++) {
      uint8_t r;
      if (restore) {
	HL_LCD_ReadData (); /* dummy read */
	r = HL_LCD_ReadData ();
	r &= ~mask;
	r |= (*restore++) & mask;
	HL_LCD_WriteData (r);
      } else {
	r = HL_LCD_ReadData ();
	*save++ = r & mask;
      }
    }
  }
  
  HL_LCD_WriteCmd (HL_LCD_INC_CTRL | 0);

  return hl_lcd_save_size (x0,y0, x1,y1);
}

/* save display area for later re-display at same position
 */
size_t hl_lcd_save (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
		    uint8_t * save, size_t size)
{
  return hl_lcd_save_restore (x0,y0, x1,y1, save,NULL,size);
}

size_t hl_lcd_restore (uint8_t x0,uint8_t y0, uint8_t x1,uint8_t y1,
		       const uint8_t * save, size_t size)
{
  return hl_lcd_save_restore (x0,y0, x1,y1, NULL,save,size);
}
