/* template source file */
/* Copyright (C) 2009 Thomas Klosa
 *               2010 Thomas Klosa
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file
 * @ingroup geXPal_lib
 * @brief  Key Scan and alphanumeric input machine implementation file
 *
 *  \par General Key Scanning Algorithm:
 *	The system will probe the key matrix using the controller internal pull-up
 *	resistors. This will avoid shorts and excessive current in case of multiple 
 *	keys beeing pressed. \n
 *	The drive end will activate the pull-up to drive the line high and sense the
 *	line state at the same time. The other end will drive active low. \n\n
 *  The Key is recognized by a 16bit scan code. It is possible to recognize more
 *  than one key, pressed simultanously by nique scan codes.\n
 *  The scan code is translated into a keyCode and stored in a buffer. 
 *  The Application can pick the inputs from the buffer later.\n\n
 *  The key input will run in background, propelled from a timer interrupt.
 *
 *	- There is no echo function
 *	- A Scan code can still bounce
 *	- A Click code is a debouned valid key scan code
 *
 *
 */


#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>

#include "expal_board.h"
#include "display.h"
#include "keys.h"
#include "async32khz.h"





/*
 *	Global KeyScanner Variables
 *
 */
	uint16_t LastScCode; 	//!< Scan Code on previous call
	uint16_t LastClickCode;	//!< Code found during last key click 
	int8_t ScanCnt;  		//!< counts how often the same key was scanned down in a consecutive order
	uint8_t CurrentKey;		//!< Key based on current clicks, could still change on next click 
	uint8_t	SplashStatus;	//!< Splash Screen Status
	uint8_t *pDisBackup;	//!< pointer to backup area for splash screen

	uint8_t KeyBuf[HL_KEY_BUFSZ];		//!< Key input buffer, Apps will peek inputs from there
	uint8_t KeyBufW; 	//!< KeyBuffer Write Count
	uint8_t KeyBufR; 	//!< KeyBuffer Read Count
	uint8_t KeyBufCC;	//!< KeyBuffer Char Count



/** @brief Matrix Hardware Init Function
 *
 *	Init key Matrix ports in a way that an active key can be detected at the COL
 *	lines as low level. This setting can also trigger an interrupt if enabled.
 */
void hlMatrixInit(void){

	// activate COL pull up resistors
	HL_KEY_COL_DDR  &= ~(HL_KEY_COL_MASK); // in
	HL_KEY_COL_PORT |=  (HL_KEY_COL_MASK); // pull

	// Set All Rows to OUT and Low
	HL_KEY_ROW_DDR_A  |=  (HL_KEY_MASK_ROW_A); // out
	HL_KEY_ROW_PORT_A &= ~(HL_KEY_MASK_ROW_A);

	HL_KEY_ROW_DDR_B  |=  (HL_KEY_MASK_ROW_B);	
	HL_KEY_ROW_PORT_B &= ~(HL_KEY_MASK_ROW_B);	

	HL_KEY_ROW_DDR_C  |=  (HL_KEY_MASK_ROW_C);	
	HL_KEY_ROW_PORT_C &= ~(HL_KEY_MASK_ROW_C);	

}

/** @brief Key Engine Init Function
 *
 *	Init has to be called once at program start
 */
void hl_Key_Init(void){

	hlMatrixInit();

	LastScCode = HL_NOKEY; 	
	LastClickCode = HL_NOKEY;	 
	ScanCnt = 0;
	CurrentKey = HL_NOKEY;
	SplashStatus = HL_KEYDIS_OFF;
	pDisBackup = NULL;
	KeyBufW=0; 
	KeyBufR=0; 
	KeyBufCC=0;

	hl_asyn_RegisterFastTimer( (&hlKeyTmTick), NULL);
}


/** @brief Scan Key Matrix
 *
 *	Scan the key matrix and return the 16bit key scan code
 *	The Scan code is determined by:
 *	ROW_ID<<  & COL_PIN & ROW_PIN
 *
 *	@return KeyScanCode
 */
uint16_t hlKeyScan(void){
	uint16_t KeyScanCode = HL_NOKEY;

	hlMatrixInit();

	/* Check if one of the COL lines is low */
	if( (HL_KEY_COL_PIN & HL_KEY_COL_MASK) != HL_KEY_COL_MASK){
		
		/* There is an active key 
		   Store COL in upper byte */
		KeyScanCode = (((~HL_KEY_COL_PIN) & HL_KEY_COL_MASK)<<8);

		/* Switch ROW's to input and activate pull up. Drive high
		   first to charge the ROW lines  */
		HL_KEY_ROW_PORT_A |= (HL_KEY_MASK_ROW_A);
		HL_KEY_ROW_PORT_B |= (HL_KEY_MASK_ROW_B);
		HL_KEY_ROW_PORT_C |= (HL_KEY_MASK_ROW_C);

		HL_KEY_ROW_DDR_A &= ~(HL_KEY_MASK_ROW_A);
		HL_KEY_ROW_DDR_B &= ~(HL_KEY_MASK_ROW_B);
		HL_KEY_ROW_DDR_C &= ~(HL_KEY_MASK_ROW_C);

		/* Turn the matrix drive low over to the COL lines*/
		HL_KEY_COL_PORT &= ~(HL_KEY_COL_MASK);
		HL_KEY_COL_DDR  |=  (HL_KEY_COL_MASK);

		_delay_us(10);  //settle lines

		/* Check ROW A */
		if((HL_KEY_ROW_PIN_A & HL_KEY_MASK_ROW_A) != HL_KEY_MASK_ROW_A){
		
			/* A key in section A is active */
			KeyScanCode |= (_BV(HL_KEY_ROW_A_ID))<<8;
			KeyScanCode |= ((~HL_KEY_ROW_PIN_A) & HL_KEY_MASK_ROW_A);
					
			return KeyScanCode; /* ---------------------------> */
		}

		/* Check ROW B */
		if((HL_KEY_ROW_PIN_B & HL_KEY_MASK_ROW_B) != HL_KEY_MASK_ROW_B){
		
			/* A key in section B is active */
			KeyScanCode |= (_BV(HL_KEY_ROW_B_ID))<<8;
			KeyScanCode |= ((~HL_KEY_ROW_PIN_B) & HL_KEY_MASK_ROW_B);
					
			return KeyScanCode; /* ---------------------------> */
		}

		/* Check ROW C */
		if((HL_KEY_ROW_PIN_C & HL_KEY_MASK_ROW_C) != HL_KEY_MASK_ROW_C){
		
			/* A key in section C is active */
			KeyScanCode |= (_BV(HL_KEY_ROW_C_ID))<<8;
			KeyScanCode |= ((~HL_KEY_ROW_PIN_C) & HL_KEY_MASK_ROW_C);
					
			return KeyScanCode; /* ---------------------------> */
		}

		/* Misdetection, COL but no ROW, probably bouncing */
		return HL_NOKEY; /* ---------------------------> */
	
	}else{
	
		/* All keys are up ------------------------------------->         */
		return KeyScanCode;
	}
}





/** @brief Display Splash Screen
 *
 *	LCD SplashScreen Functions to display the multiKey status.
 *	The function is using the location definition from the header file
 *
 * @param[in] keys  Vector of selectable keys
 * @param[in] Akt   Position of activated key (will be displayed inverted)
 */

void hlShowKeySplash(uint8_t *keys, uint8_t Akt){

	char s[8];
	s[7] = 0;

	if(SplashStatus == HL_KEYDIS_OFF){
		// allocate mem and backup splash area when off
		// Splash Area backup
		size_t BackupSize = hl_lcd_save_size (	HL_KEYDIS_X,HL_KEYDIS_Y, \
												HL_KEYDIS_X+HL_KEYDIS_W, \
												HL_KEYDIS_Y+HL_KEYDIS_H);
		pDisBackup = malloc (BackupSize);

		if (pDisBackup != NULL){
			// Mark Status as ON
			SplashStatus = HL_KEYDIS_ON;
			hl_lcd_save (	HL_KEYDIS_X,HL_KEYDIS_Y, \
							HL_KEYDIS_X+HL_KEYDIS_W, \
							HL_KEYDIS_Y+HL_KEYDIS_H, pDisBackup, BackupSize);
		}else{
			// TODO:  Exception Handler
		  printf_P(PSTR("\n ERROR: malloc failed in key splash routine\n"));
		}
	}

	if(SplashStatus == HL_KEYDIS_ON){
		// Splash Area is available

		// Splash Clean and draw Frame
		hl_lcd_set_rectangle(	HL_KEYDIS_X,HL_KEYDIS_Y, \
								HL_KEYDIS_X+HL_KEYDIS_W, \
								HL_KEYDIS_Y+HL_KEYDIS_H, op_clear);
		// top
		hl_lcd_set_rectangle(	HL_KEYDIS_X,HL_KEYDIS_Y, \
								HL_KEYDIS_X+HL_KEYDIS_W, \
								HL_KEYDIS_Y+HL_KEYDIS_FR_W, op_set);
		// bottom
		hl_lcd_set_rectangle(	HL_KEYDIS_X,\
								HL_KEYDIS_Y+HL_KEYDIS_H-HL_KEYDIS_FR_W, \
								HL_KEYDIS_X+HL_KEYDIS_W, \
								HL_KEYDIS_Y+HL_KEYDIS_H, op_set);
		// left
		hl_lcd_set_rectangle(	HL_KEYDIS_X,HL_KEYDIS_Y, \
								HL_KEYDIS_X+HL_KEYDIS_FR_W, \
								HL_KEYDIS_Y+HL_KEYDIS_H, op_set);
		// right
		hl_lcd_set_rectangle(	HL_KEYDIS_X+HL_KEYDIS_W-HL_KEYDIS_FR_W, \
								HL_KEYDIS_Y, \
								HL_KEYDIS_X+HL_KEYDIS_W, \
								HL_KEYDIS_Y+HL_KEYDIS_H, op_set);

		// Draw Available Keys 
		  sprintf_P(s, PSTR("%c %c %c %c"), \
				pgm_read_byte(&keys[0]),\
				pgm_read_byte(&keys[1]), \
				pgm_read_byte(&keys[2]), \
				pgm_read_byte(&keys[3]) );
		hl_lcd_drawstr (HL_KEYDIS_X+HL_KEYDIS_TO_X, HL_KEYDIS_Y+HL_KEYDIS_TO_Y, s);

		// Invert active Key
		hl_lcd_set_rectangle(	HL_KEYDIS_X+HL_KEYDIS_TO_X+(Akt*2*HL_KEYDIS_FNT_W)-2,\
								HL_KEYDIS_Y+HL_KEYDIS_TO_Y-2, \
								HL_KEYDIS_X+HL_KEYDIS_TO_X+(Akt*2*HL_KEYDIS_FNT_W)+HL_KEYDIS_FNT_W+1, \
								HL_KEYDIS_Y+HL_KEYDIS_TO_Y+HL_KEYDIS_FNT_H+1, \
								op_invert);
	}
}


/** @brief Restore the display splash screen area.
 *
 *	- Restore the display splash screen area.
 *	- return allocated memory
 *
 */

void hlRestoreKeySplash(void){

	if (SplashStatus == HL_KEYDIS_ON){

		// restore splash area
		size_t BackupSize = hl_lcd_save_size (HL_KEYDIS_X,HL_KEYDIS_Y, HL_KEYDIS_X+HL_KEYDIS_W, HL_KEYDIS_Y+HL_KEYDIS_H);

		if (pDisBackup != NULL){

			hl_lcd_restore (HL_KEYDIS_X,HL_KEYDIS_Y, HL_KEYDIS_X+HL_KEYDIS_W, HL_KEYDIS_Y+HL_KEYDIS_H, pDisBackup, BackupSize);
		
			// return allocated memory
			free(pDisBackup);
	
		}else{
			// TODO:  Exception Handler
		  printf_P(PSTR("\n ERROR: nullpointer when doing free() in key splash routine\n"));
		}
	
	// Stat is OFF now
	SplashStatus = HL_KEYDIS_OFF;

	}
}


/** \verbatim Multi Push Code Table
 *
 * Definition Table for Multi push key return codes
 *
 */
uint8_t HlKeyNum1[4] PROGMEM  = {HL_KEY_1, HL_KEY_A, HL_KEY_B, HL_KEY_C}; 
uint8_t HlKeyNum2[4] PROGMEM  = {HL_KEY_2, HL_KEY_D, HL_KEY_E, HL_KEY_F};
uint8_t HlKeyNum3[4] PROGMEM  = {HL_KEY_3, HL_KEY_G, HL_KEY_H, HL_KEY_I};
uint8_t HlKeyNum4[4] PROGMEM  = {HL_KEY_4, HL_KEY_J, HL_KEY_K, HL_KEY_L};
uint8_t HlKeyNum5[4] PROGMEM  = {HL_KEY_5, HL_KEY_M, HL_KEY_N, HL_KEY_O};
uint8_t HlKeyNum6[4] PROGMEM  = {HL_KEY_6, HL_KEY_P, HL_KEY_Q, HL_KEY_R};
uint8_t HlKeyNum7[4] PROGMEM  = {HL_KEY_7, HL_KEY_S, HL_KEY_T, HL_KEY_U};
uint8_t HlKeyNum8[4] PROGMEM  = {HL_KEY_8, HL_KEY_V, HL_KEY_W, HL_KEY_X};
uint8_t HlKeyNum9[4] PROGMEM  = {HL_KEY_9, HL_KEY_Y, HL_KEY_Z, HL_KEY_SZ};
uint8_t HlKeyNum0[4] PROGMEM  = {HL_KEY_0, HL_KEY_AE, HL_KEY_OE, HL_KEY_UE};
uint8_t HlKeyStar[4] PROGMEM  = {HL_KEY_STAR, HL_KEY_DOT, HL_KEY_KOMMA, HL_KEY_SLASH};
uint8_t HlKeyPound[4]PROGMEM  = {HL_KEY_POUND, HL_KEY_MINUS, HL_KEY_PLUS, HL_KEY_EQL};
/* \endverbatim */



/** @brief Decode ScanCodes into KeyKodes
 *
 *	Decoder State Machine to determine inputs from scan codes
 *	This function will also return entries from multi function keys
 *
 * @param[in] ScCode KeyScanCode
 * @param[in] LastClickCode ScanCode of last key click
 * @return Key Code
 */
uint8_t hlKeyDecode(uint16_t ScCode, uint16_t LastClickCode){

	static uint8_t RptCnt;  		// counts number of key klicks
	uint8_t KeyCode;
	uint8_t *mkeys;

	mkeys = NULL;

	if(ScCode == LastClickCode){
		// same key recognized again
		RptCnt++;
		if (RptCnt > HL_KEY_MAX_KEY_FUNCT){
			// jump back to first key code
			RptCnt = 0;
		}	
	}else{
		// different key, start with first key code
		RptCnt = 0;	
	}

	switch(ScCode){
	// Display soft keys
	case HL_SCAN_SOFT1: KeyCode = HL_KEY_SOFT_1;  break;
	case HL_SCAN_SOFT2: KeyCode = HL_KEY_SOFT_2;  break; 
	case HL_SCAN_SOFT3: KeyCode = HL_KEY_SOFT_3;  break;
	case HL_SCAN_SOFT4: KeyCode = HL_KEY_SOFT_4;  break;
	case HL_SCAN_SOFT5: KeyCode = HL_KEY_SOFT_5;  break;
	case HL_SCAN_SOFT6: KeyCode = HL_KEY_SOFT_6;  break;
	case HL_SCAN_SOFT7: KeyCode = HL_KEY_SOFT_7;  break;
	case HL_SCAN_SOFT8: KeyCode = HL_KEY_SOFT_8;  break;
	case HL_SCAN_SOFT9: KeyCode = HL_KEY_SOFT_9;  break;
	case HL_SCAN_SOFT10: KeyCode = HL_KEY_SOFT_A; break;
	case HL_SCAN_SOFT11: KeyCode = HL_KEY_SOFT_B; break;

	// left thumb keys
	case HL_SCAN_LEFT1: KeyCode = HL_KEY_LEFT_1; break;
	case HL_SCAN_LEFT2: KeyCode = HL_KEY_LEFT_2; break;
	case HL_SCAN_LEFT3: KeyCode = HL_KEY_LEFT_3; break;
	case HL_SCAN_LEFT4: KeyCode = HL_KEY_LEFT_4; break;

	// right thumb keys
	case HL_SCAN_RIGHT1: KeyCode = HL_KEY_RIGHT_1; break;
	case HL_SCAN_RIGHT2: KeyCode = HL_KEY_RIGHT_2; break;
	case HL_SCAN_RIGHT3: KeyCode = HL_KEY_RIGHT_3; break;
	case HL_SCAN_RIGHT4: KeyCode = HL_KEY_RIGHT_4; break;

	// Number Key Pad
	case HL_SCAN_NUM0 :	KeyCode = pgm_read_byte (&(HlKeyNum0 [RptCnt])); mkeys=HlKeyNum0; break;
	case HL_SCAN_NUM1 :	KeyCode = pgm_read_byte (&(HlKeyNum1 [RptCnt])); mkeys=HlKeyNum1; break;
	case HL_SCAN_NUM2 :	KeyCode = pgm_read_byte (&(HlKeyNum2 [RptCnt])); mkeys=HlKeyNum2; break;
	case HL_SCAN_NUM3 :	KeyCode = pgm_read_byte (&(HlKeyNum3 [RptCnt])); mkeys=HlKeyNum3; break;
	case HL_SCAN_NUM4 :	KeyCode = pgm_read_byte (&(HlKeyNum4 [RptCnt])); mkeys=HlKeyNum4; break;
	case HL_SCAN_NUM5 :	KeyCode = pgm_read_byte (&(HlKeyNum5 [RptCnt])); mkeys=HlKeyNum5; break;
	case HL_SCAN_NUM6 :	KeyCode = pgm_read_byte (&(HlKeyNum6 [RptCnt])); mkeys=HlKeyNum6; break;
	case HL_SCAN_NUM7 :	KeyCode = pgm_read_byte (&(HlKeyNum7 [RptCnt])); mkeys=HlKeyNum7; break;
	case HL_SCAN_NUM8 :	KeyCode = pgm_read_byte (&(HlKeyNum8 [RptCnt])); mkeys=HlKeyNum8; break;
	case HL_SCAN_NUM9 :	KeyCode = pgm_read_byte (&(HlKeyNum9 [RptCnt])); mkeys=HlKeyNum9; break;
	case HL_SCAN_STAR :	KeyCode = pgm_read_byte (&(HlKeyStar [RptCnt])); mkeys=HlKeyStar; break;
	case HL_SCAN_POUND:	KeyCode = pgm_read_byte (&(HlKeyPound [RptCnt])); mkeys=HlKeyPound; break;

	default: // unrecognized key or multiple key press 
			 KeyCode = HL_NOKEY;   
			 break;
	}
	// Show Splash if MultiKey
	if(KeyCode < HL_KEY_MULTI){
		hlShowKeySplash(mkeys, RptCnt);
	}
	return KeyCode;
}

/** @brief Key debounce state machine
 *
 *	Debounce and multi click counters
 *	In case of multi key, the user can choose other entries with consecutive clicks
 *	like 1 - a - b - c - 1...
 *	the final value will be returned after timeout or when an other key is detected as down.
 *
 *	- There is no echo function
 *	- A Scan code can still bounce
 *	- A Click code is a debouned valid key scan code
 *
 *  @param ScCode 16bit Scancode, returned from hlKeyScan()
 *
 *	@return HL_NOKEY or a valid keycode
 *
 *
 *	\note :-)  Die Funktion gefllt mir schon nicht mehr bevor sie fertig ist..  :-(
 *	            Vielleicht doch besser auf die gute alte statemachine setzen.
 */
uint8_t  hlKeyDebounce(uint16_t ScCode){

	uint8_t FinalKey;   		// Final key, after timeout or click to an other key
	FinalKey = HL_NOKEY;  

	if( (ScCode != HL_NOKEY) && (ScanCnt != HL_KEY_DONE)){

		if(ScCode == LastScCode){		
			// Last Key still down
			if( ScanCnt == HL_KEY_DONE){			
				FinalKey = HL_NOKEY;  	

			}else{
				// valid unhandled key is down
				ScanCnt++;
				if(ScanCnt > HL_KEY_DEBOUNCE_CYCLES){
					// Valid key detected
					ScanCnt = HL_KEY_DONE;
					// store key as current entry until wait time over or other key pressed
					CurrentKey = hlKeyDecode(ScCode, LastClickCode);
					LastClickCode = ScCode;
					if(CurrentKey >= HL_KEY_MULTI){
						// return key when single click key
						FinalKey = CurrentKey;
						CurrentKey = HL_NOKEY;
					}// return multi keys when timeout over or other key is down
				}
			}		
		}else{
			// different key start fresh
			ScanCnt = 0;
			LastScCode = ScCode;

			// clear last click if new key other than LastClick
			// previous state can be NO_KEY or something wired in case of multiple keys
			if (LastClickCode != ScCode){
				// user did move over to an other key

				LastClickCode = HL_NOKEY;

				// return previous multi key
				FinalKey = CurrentKey;
				CurrentKey = HL_NOKEY;

			}
		}
	}else{
		// all keys up, get out here quick
		LastScCode = ScCode;
		// Count up for multiKey Timeout and to delete KEY_DONE
		if (ScCode == HL_NOKEY){ 
			ScanCnt++;
		}

		if(ScanCnt > HL_KEY_WAIT_ALT_FKT){
			// wait time is over, delete klick history
			LastClickCode = HL_NOKEY;
			ScanCnt = 0;
			// return multi key
			FinalKey = CurrentKey;
			CurrentKey = HL_NOKEY;
		}
	}

	if (CurrentKey == HL_NOKEY){
		hlRestoreKeySplash();
	}
	return FinalKey;  
}


/** @brief Timer call to drive the key scan
 *
 *	Function to be called by a timer with the polling interval frequency
 */
void hlKeyTmTick(void){

	uint16_t ksc;
	uint8_t key;

	ksc = hlKeyScan();
	//sprintf(s, "\nKeyScanCode: %4X ", ksc);
	//printf(s);
	//hl_lcd_drawstr (1*8, 45, s);

	key = hlKeyDebounce(ksc);
	if(key != HL_NOKEY){

		//sprintf(s, "\nKeyCode: %2X ", key);
		//printf(s);
		//hl_lcd_drawstr (1*8, 55, s);

		// KeyBuffer will always contain the last 8 keys
		KeyBuf[KeyBufW]=key;
		KeyBufW++;
		KeyBufCC++;
		if(KeyBufW >= HL_KEY_BUFSZ){
			KeyBufW=0;
		}
		if (KeyBufCC > HL_KEY_BUFSZ){
			KeyBufCC--;
			// Buffer is full, forget the oldest entry
			KeyBufR++;
			if(KeyBufR >= HL_KEY_BUFSZ){
				KeyBufR=0;
			}		
		}
	}
}


/** @brief  peek a key from the key buffer
 *
 * Get input keys from the character buffer
 *
 * @param[out] Key Pointer to store the key value 
 * @return returns number of keys in buffer or 0 if empty
 */
uint8_t hlGetKey(uint8_t *Key){

	if(KeyBufCC > 0){
		// There is something in the buffer
		*Key = KeyBuf[KeyBufR];

		KeyBufCC--;

		KeyBufR++;
		if(KeyBufR >= HL_KEY_BUFSZ){
			KeyBufR=0;
		}		
		return KeyBufCC;   // ------------------------------------->
	}else{

		return HL_NOKEY;   // ------------------------------------->
	}
}

