/* template source file */
/* Copyright (C) 2009 Thomas Klosa, Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include "lh155.h"
#include "expal_board.h"

/* needed for cycle-correct delays */
#include "delay_x_gcc.h"

/* delay for read or write pulse                      */
/* parameter:                                         *
 *  rd   if !=0, we need the delay for a read cycle   *
 *       if ==0, we need the delay for a write cycle  *
 */
static void hl_lcd_clk_delay (uint8_t rd)
  __attribute__((always_inline));
static inline void
hl_lcd_clk_delay (uint8_t rd)
{
  /* port access itself needs at least two clock cycles */
  if (rd) {
    /* we need a delay of 270 ns */
    _delay_ns (270 - 1.5e9/F_CPU);
  } else {
    /* we need a delay of 100 ns */
    _delay_ns (100 - 1.5e9/F_CPU);
  }
}

static uint8_t hl_lcd_transfer (uint8_t reg, uint8_t read, uint8_t data)
  __attribute__((always_inline));
static inline uint8_t hl_lcd_transfer (uint8_t reg, uint8_t read,
				       uint8_t data)
{
  /* setup data port */
  if (read) {
    HL_LCD_DATA_DDR = 0x00;
    HL_LCD_DATA_PORT = 0x00; /* disable pull-ups */
  } else {
    HL_LCD_DATA_PORT = data;
    HL_LCD_DATA_DDR = 0xFF;
  }

  /* setup register/data signal */
  if (reg) {
    HL_LCD_D_R_PORT |= _BV(HL_LCD_D_R);
  } else {
    HL_LCD_D_R_PORT &= ~_BV(HL_LCD_D_R);
  }

  /* activate CS */
  HL_LCD_CS_PORT &= ~_BV(HL_LCD_CS);

  /* start cycle */
  if (read) {
    HL_LCD_RD_PORT &= ~_BV(HL_LCD_RD);
  } else {
    HL_LCD_WR_PORT &= ~_BV(HL_LCD_WR);
  }
  
  hl_lcd_clk_delay (read);

  /* transfer data and stop cycle */
  if (read) {
    /*_delay_us (500);*/
    data = HL_LCD_DATA_PIN;
    HL_LCD_RD_PIN = _BV(HL_LCD_RD);
  } else {
    HL_LCD_WR_PIN = _BV(HL_LCD_WR);
  }

  /* deactivate CS */
  HL_LCD_CS_PIN = _BV(HL_LCD_CS);

  return data;
}

/* write a command into control register */
void HL_LCD_WriteCmd (uint8_t cmd)
{
  hl_lcd_transfer (1, 0, cmd);
}

void HL_LCD_WriteData (uint8_t Data)
{
  hl_lcd_transfer (0, 0, Data);
}

uint8_t HL_LCD_ReadData (void)
{
  uint8_t data;

  data = hl_lcd_transfer (0, 1, 0);

  return data;
}


void HL_LCD_SetX (uint8_t XAdr)
{
  HL_LCD_WriteCmd( (HL_LCD_PARAM_MASK & XAdr) | HL_LCD_X_ADR_L_CMD );
  HL_LCD_WriteCmd( (HL_LCD_PARAM_MASK & (XAdr >> 4)) | HL_LCD_X_ADR_H_CMD);
}

void HL_LCD_SetY (uint8_t YAdr)
{
  HL_LCD_WriteCmd( (HL_LCD_PARAM_MASK & YAdr) | HL_LCD_Y_ADR_L_CMD );
  HL_LCD_WriteCmd( (HL_LCD_PARAM_MASK & (YAdr >> 4)) | HL_LCD_Y_ADR_H_CMD );
}

void HL_LCD_WriteByte (uint8_t XAdr, uint8_t YAdr, uint8_t Data){
  HL_LCD_SetX (XAdr);
  HL_LCD_SetY (YAdr);
  HL_LCD_WriteData (Data);
}

void HL_LCD_Init (void){

	// CTRL line init
	HL_LCD_RESET_PORT |= _BV(HL_LCD_RESET);
	HL_LCD_CS_PORT |= _BV(HL_LCD_CS);
	HL_LCD_D_R_PORT |= _BV(HL_LCD_D_R);
	HL_LCD_WR_PORT |= _BV(HL_LCD_WR);
	HL_LCD_RD_PORT |= _BV(HL_LCD_RD);

	HL_LCD_RESET_DDR |= _BV(HL_LCD_RESET);
	HL_LCD_CS_DDR |= _BV(HL_LCD_CS);
	HL_LCD_D_R_DDR |= _BV(HL_LCD_D_R);
	HL_LCD_WR_DDR |= _BV(HL_LCD_WR);
	HL_LCD_RD_DDR |= _BV(HL_LCD_RD);

	// Data port init
	HL_LCD_DATA_DDR = 0xFF;  // out
	HL_LCD_DATA_PORT= 0x00;

	// Display Reset
	HL_LCD_RESET_PORT &= ~(_BV(HL_LCD_RESET));
	_delay_us(15); /* data sheet, page 41 */
	HL_LCD_RESET_PORT |= _BV(HL_LCD_RESET);
	
	// Display init sequence
	HL_LCD_WriteCmd (HL_LCD_CLR_RE);

	HL_LCD_WriteCmd ( HL_LCD_START_LINE_L );
	HL_LCD_WriteCmd ( HL_LCD_START_LINE_H );

	HL_LCD_WriteCmd ( HL_LCD_ALT_LINE_L | 1 );  // line 1
	HL_LCD_WriteCmd ( HL_LCD_ALT_LINE_H );

	HL_LCD_WriteCmd ( HL_LCD_DISP_CTRL_1 | 1 ); // display on
	HL_LCD_WriteCmd ( HL_LCD_DISP_CTRL_2 ); /* swap off */

	HL_LCD_WriteCmd ( HL_LCD_INC_CTRL | HL_LCD_INC_CTRL_AXI );

	HL_PWR_VLCD_DDR  |= _BV(HL_PWR_VLCD);		// 15V LCD Voltage on
	HL_PWR_VLCD_PORT |= _BV(HL_PWR_VLCD);
	HL_LCD_WriteCmd ( HL_LCD_PWR_CTRL_1 | 2 );	// PowerOn

	HL_LCD_WriteCmd ( HL_LCD_PWR_CTRL_2 | 0x0E );	// el volume off

	HL_LCD_WriteCmd (HL_LCD_SET_RE);

	HL_LCD_WriteCmd ( HL_LCD_PWR_CTRL_3 );		// duty cycle 1/64

	HL_LCD_WriteCmd (HL_LCD_CLR_RE);

}

