/* implement low-frequency timer */
/* Copyright (C) 2009,2010 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#if 0
/* the macros in util/atomic.h introduce "memory barriers" which are
 * not necessary here
 */
#include <util/atomic.h>
#endif

#include "rtc.h"
#include "cb_list.h"

/* activate low-frequency oscillator
 */
void hl_rtc_osc_init (void)
{
  /* disable interrupts from timer during start-up */
  TIMSK2 = 0;

  /* stop timer before enabling asynchronous operation */
  TCCR2B &= ~(_BV(CS22)|_BV(CS21)|_BV(CS20));
  while (ASSR & _BV(TCR2BUB));

  /* enable asynchronous operation with quartz oscillator */
  ASSR = _BV(AS2);

  /* wait until previous TCNT update finishes */
  while (ASSR & _BV(TCN2UB));

  /* reset counter */
  TCNT2 = 0;

}

#define RTC_SCALE (80)
#define RTC_PRESCALE (8)
/* Prescaler = 8 */
#define RTC_PRESCALER_CONTROL (0*_BV(CS22)|_BV(CS21)|0*_BV(CS20))

/* ensure that UXSEC_PER_TICK is an integer value! */
#define Ux (4L)
#define UxSEC_HZ (Ux*1000000)
/* we assume F_RTC_OSC % RTC_PRESCALE == 0 */
#define UxSEC_PER_TICK (UxSEC_HZ*RTC_SCALE/(F_RTC_OSC/RTC_PRESCALE))

/* initialize low-frequency timer */
void hl_rtc_init (void)
{
  /* wait until previous TCCR2x updates finish */
  while (ASSR & (_BV(TCN2UB)|_BV(OCR2AUB)|_BV(TCR2AUB)|_BV(TCR2BUB)));

  /* set counter register to known state */
  TCNT2 = 0;  OCR2A = RTC_SCALE-1;

  TCCR2A = 0
    /* | _BV(COM2A1) | _BV(COM2A0) */ /* disconnect OC2A */
    /* | _BV(COM2B1) | _BV(COM2B0) */ /* disconnect OC2B */
    | _BV(WGM21) /* | _BV(WGM20) */ /* CTC mode */
    ;
  TCCR2B = 0
    /* | _BV(FOC2A) | _BV(FOC2B) */ /* no Force Output Compare */
    /* | _BV(WGM22) */ /* CTC mode */
    | RTC_PRESCALER_CONTROL
    ;
  
  TIFR2 = _BV(OCF2A); /* clear OV interrupt flag */
  TIMSK2 = 0
    /* | _BV(OCIE2B) */ | _BV(OCIE2A)
    /* | _BV(TOIE2) */
    ;
}

/* variables for time-keeping */

/* ticks is only updated by timer interrupt routine and this is the
 * only thing the interrupt routine does
 */
static uint16_t volatile ticks /*=0*/; /* init done by BSS clear */

/* atomically get the value of the ticks variable */
static inline uint16_t get_ticks (void)
{
  uint16_t t;
  uint8_t sreg = SREG;
  cli ();
  t = ticks;
  SREG = sreg;
  return t;
}

ISR(TIMER2_COMPA_vect)
{
  ticks++;
}


cb_list_t rtc_1s_list;
cb_list_t rtc_fast_list;

void hl_asyn_Register1sTimer(cb_function_t fkt2call, void *arg)
{
  cb_register_call (&rtc_1s_list, fkt2call, arg);
}

void hl_asyn_RegisterFastTimer(cb_function_t fkt2call, void *arg)
{
  cb_register_call (&rtc_1s_list, fkt2call, arg);
}

static uint16_t old_ticks /*=0*/;

/* 1/Ux micro-second counter
 * not guaranteed to be <Ux*1000000L
 */
static uint32_t time_uxsec;

/* seconds since startup
 * for correct time, add time_usec/(Ux*1000000L)
 */
static uint32_t time_sec;
static uint8_t time_sec_disabled;

static void normalize_uxsec (void)
{
  while (time_uxsec >= UxSEC_HZ) {
    time_uxsec -= UxSEC_HZ;
    time_sec++;
  }
}

static void ticks_to_uxsec (uint16_t n)
{
  uint32_t uincr = (uint32_t)n*UxSEC_PER_TICK;
  uint32_t time_uxsec_new = time_uxsec + uincr;
  if (time_uxsec_new < time_uxsec) {
    /* handle overflow */
    uint32_t sincr = uincr/UxSEC_HZ;
    sincr += time_uxsec/UxSEC_HZ;
    time_sec += sincr;
    time_uxsec_new -= sincr*UxSEC_HZ;
    while (sincr--)
      callback_iterate (rtc_1s_list);
  } else if (!time_sec_disabled && time_uxsec_new >= UxSEC_HZ) {
    time_sec++;
    time_uxsec_new -= UxSEC_HZ;
    callback_iterate (rtc_1s_list);
  }
  {
    for (; uincr >= UxSEC_HZ/20000 /*20ms*/; uincr -= UxSEC_HZ/20000)
      callback_iterate (rtc_fast_list);
  }
  time_uxsec = time_uxsec_new;
  old_ticks += n;
}

/* update time variables from ticks
 * ensure this gets called before ticks crosess old_ticks
 * (i.e. at intervals shorter than 1024/Ux seconds)
 */
static void hl_rtc_update_time (void)
{
  while (1) {
    uint16_t t;
    t = get_ticks ();
    if (t==old_ticks) break;
    ticks_to_uxsec (t-old_ticks);
  }
}

hl_time_t hl_get_current_time (void)
{
  hl_time_t ret;
  hl_rtc_update_time ();
  ret.seconds = time_sec;
  ret.useconds = time_uxsec/Ux;
  while (ret.useconds >= UxSEC_HZ/Ux) {
    /* ticks_to_usec does only guarantee that time_uxsec < Ux*USEC_HZ */
    ret.useconds -= UxSEC_HZ/Ux;
    ret.seconds++;
  }
  /* debug aid */
  ret.ticks = get_ticks ();
  return ret;
}

/* enter power-save mode until next timer overflow
 * or other wake-up event
 */
void hl_sleep_for_tick (uint16_t ms)
{
  uint32_t end_uxsec;
  uint8_t tsd_save;

  hl_rtc_update_time ();
  set_sleep_mode (SLEEP_MODE_IDLE); /* keep USART running */
  cli ();
  tsd_save = time_sec_disabled;
  time_sec_disabled = 1;
  end_uxsec = time_uxsec + ms*Ux*1000L;
  sleep_enable ();
  while (time_uxsec < end_uxsec) {
    uint16_t t;
    sei ();
    sleep_cpu ();
    
    while ((t = get_ticks ()) != old_ticks)
      ticks_to_uxsec (t - old_ticks);
  }
  sleep_disable ();
  time_sec_disabled = tsd_save;
  sei ();
  ticks_to_uxsec (0);
}
