/* USART access */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*!\file
 * \ingroup geXPal_lib
 * \brief generic USART access implementation
 */

#include "usarts.h"
#include <avr/interrupt.h>
#include "expal_board.h"
#include "usarts_int.h"

#define USART_DEF USART_DEF_CTRL

#include "usarts_all.h"

/* baud rate definitions */
#define BAUD_TOL (1)
static struct serial_param_s {
  uint16_t baud;
  uint16_t rate;
  uint8_t use_2x;
} serial_param [] PROGMEM = {
#include "baud-params.h"
  {0, 0, 0} /* END MARKER */
};

uint16_t usart_set_baudrate (usart_control_t * uc, uint16_t baud)
{
  const struct serial_param_s *p;
  uint16_t bd;

  for (p = serial_param; (bd = pgm_read_word(&p->baud), bd && (bd != baud)); p++);

  if (bd) {
    uint16_t rate = pgm_read_word (&p->rate);
    UBRRnH(uc) = rate >> 8;
    UBRRnL(uc) = rate & 0xFF;
    if (pgm_read_byte (&p->use_2x))
      UCSRnA(uc) |= _BV(U2X0);
    else
      UCSRnA(uc) &= ~_BV(U2X0);

    return baud;
  }
  return 0;
}

usart_control_t * usart_init (hlog_usart_t which,
			      fifo_item_t * txbuf, fifo_size_t txsize,
			      fifo_item_t * rxbuf, fifo_size_t rxsize)
{
  usart_control_t * uc;
  uint8_t ctl;

  switch (which) {
  case usart_ext:
    uc = &usart_control_3;
    ctl = HL_EX232_CTL_PORT & ~(_BV(HL_EX232_EN) | _BV(HL_EX232_FRCON));
    HL_EX232_CTL_PORT = ctl;
    break;
  case usart_mod:
    uc = &usart_control_2;
    break;
  default:
    return 0;
  }

  /* enable power */
  *(uc->prr) &= ~(uc->prusart);

  fifo_init (&uc->txbuf, txbuf, txsize);
  fifo_init (&uc->rxbuf, rxbuf, rxsize);

  /* set frame format: 8N1 */
  UCSRnC(uc) = _BV(UCSZ01) | _BV(UCSZ00);

  return uc;
}

void usart_disable (usart_control_t * uc)
{
  UCSRnB(uc) &= ~(_BV(RXCIE0)|_BV(TXCIE0)|_BV(UDRIE0)
		  |_BV(RXEN0)|_BV(TXEN0));
}

void usart_poweroff (usart_control_t * uc)
{
  *(uc->prr) |= (uc->prusart);
  if (uc == &usart_control_3) {
    /* external */
    HL_EX232_CTL_PORT |= _BV(HL_EX232_EN);
  }
}

void usart_enable (usart_control_t * uc)
{
  UCSRnA(uc) &= _BV(U2X0); /* reset all bits except U2X (double baudrate) */
  UCSRnB(uc) |= (fifo_get_buffer(&uc->rxbuf) ? (_BV(RXCIE0)|_BV(RXEN0)) : 0)
    | (fifo_get_buffer(&uc->txbuf) ? (_BV(UDRIE0)|_BV(TXEN0)) : 0);
}

