/* Displaytech LCD module driver */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "lcd-displaytech.h"

#include "protoconfig.h"
#include <util/delay.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "protostatus.h"

/* port definitions */
#define PORT_LCD_DT_C  PORTD
#define PORT_LCD_DT_D  PORTC
#define DDR_LCD_DT_C   DDRD
#define DDR_LCD_DT_D   DDRC
#define PIN_LCD_DT_D   PINC

#define LCD_DATA _BV(PORTD7)
#define LCD_READ _BV(PORTD6)
#define LCD_ENABLE _BV(PORTD5)
#define DD_LCD_DATA   DDD7
#define DD_LCD_READ   DDD6
#define DD_LCD_ENABLE DDD5

/* command definitions */
#define LCD_CLEAR          (0b00000001)

#define LCD_RETURN_HOME    (0b00000010)

#define LCD_ENTRY_MODE     (0b00000100)
# define LCD_CURSOR_INCR   (0b00000010)
# define LCD_DISP_SHIFT    (0b00000001)

#define LCD_DISPLAY_OFF    (0b00001000)
# define LCD_DISPLAY_ONOFF (0b00000100)
# define LCD_CURSOR_ON     (0b00000010)
# define LCD_BLINK_ON      (0b00000001)
#define LCD_DISPLAY_ON (LCD_DISPLAY_OFF | LCD_DISPLAY_ONOFF)

#define LCD_SHIFT          (0b00010000)
# define LCD_SHIFT_DISP    (0b00001000)
# define LCD_SHIFT_RIGHT   (0b00000100)

#define LCD_SET_FUNCTION   (0b00100000)
# define LCD_8BIT          (0b00010000)
# define LCD_2LINE         (0b00001000)
# define LCD_5X10          (0b00000100)

#define LCD_SET_CGADDR     (0b01000000)

#define LCD_SET_DDADDR     (0b10000000)

#define LCD_BUSY_FLAG      (0b10000000)

#define DEBUG_LCD

static inline void lcd_clk_delay (uint8_t phase)
  __attribute__((always_inline));
static void
lcd_clk_delay (uint8_t phase)
{
#ifdef DEBUG_LCD
  _delay_ms(40);
  if (phase) {
    _delay_ms(40);
    proto_status(PIN_LCD_DT_D, STATUS_LCD);
  }
#else
  /* we need a delay of 250 ns */
#if F_CPU >  4000000UL
  __asm__ volatile ("nop");
#if F_CPU >  8000000UL
  __asm__ volatile ("nop");
#if F_CPU > 12000000UL
  __asm__ volatile ("nop");
#endif
#endif
#else
  /* for clock frequences <= 4MHZ, 1 clock cycle is enough */
#endif
#endif
}

static void
lcd_put_cmd (uint8_t cmd)
{
  lcd_clk_delay (0);
  DDR_LCD_DT_D = 0xFF;
  PORT_LCD_DT_C &= ~(LCD_DATA|LCD_READ);
  PORT_LCD_DT_C |= LCD_ENABLE;
  PORT_LCD_DT_D = cmd;
  lcd_clk_delay (1);
  PORT_LCD_DT_C &= ~LCD_ENABLE;
}

static void
lcd_put_data (uint8_t data)
{
  lcd_clk_delay (0);
  DDR_LCD_DT_D = 0xFF;
  PORT_LCD_DT_C = (PORT_LCD_DT_C & ~LCD_READ) | LCD_DATA;
  PORT_LCD_DT_C |= LCD_ENABLE;
  PORT_LCD_DT_D = data;
  lcd_clk_delay (1);
  PORT_LCD_DT_C &= ~LCD_ENABLE;
}

static uint8_t
lcd_get_status (void)
{
  uint8_t data;
  PORT_LCD_DT_D = 0x00;
  DDR_LCD_DT_D = 0x00;
  lcd_clk_delay (0);
  PORT_LCD_DT_C = (PORT_LCD_DT_C & ~LCD_DATA) | LCD_READ;
  PORT_LCD_DT_C |= LCD_ENABLE;
  lcd_clk_delay (1);
  data = PIN_LCD_DT_D;
  PORT_LCD_DT_C &= ~LCD_ENABLE;
  return data;
}

static uint8_t
lcd_dt_busy (void)
{
#if 1
  return lcd_get_status() & LCD_BUSY_FLAG;
#else
  _delay_us (50);
  return 0;
#endif
}

static void
lcd_dt_reset (void)
{
#if 1
  _delay_ms (15);
  proto_status (2, STATUS_CONT);
  lcd_put_cmd (LCD_SET_FUNCTION | LCD_8BIT);
  _delay_ms(4.1);
  proto_status (3, STATUS_CONT);
  lcd_put_cmd (LCD_SET_FUNCTION | LCD_8BIT);
  _delay_us (100);
  proto_status (4, STATUS_CONT);
  lcd_put_cmd (LCD_SET_FUNCTION | LCD_8BIT);
#else
  _delay_ms(200);
  proto_status (4, STATUS_CONT);
#endif
  while (lcd_dt_busy ());
  proto_status (5, STATUS_CONT);
  lcd_put_cmd (LCD_SET_FUNCTION | LCD_8BIT | LCD_2LINE);
  while (lcd_dt_busy ());
  proto_status (6, STATUS_CONT);
  lcd_put_cmd (LCD_DISPLAY_OFF);
  while (lcd_dt_busy ());
  proto_status (7, STATUS_CONT);
  lcd_put_cmd (LCD_CLEAR);
  while (lcd_dt_busy ());
  proto_status (8, STATUS_CONT);
  lcd_put_cmd (LCD_ENTRY_MODE | LCD_CURSOR_INCR);
}

uint8_t
lcd_disp_init (void)
{
  static PROGMEM char msg[] = "huhu!";
  prog_char * s = msg;
  char c;

  DDR_LCD_DT_C |= _BV(DD_LCD_DATA) | _BV(DD_LCD_READ) | _BV(DD_LCD_ENABLE);
  DDR_LCD_DT_D = 0xFF; /* output */
  
  lcd_dt_reset ();
  while (lcd_dt_busy ());
  proto_status (9, STATUS_CONT);
  lcd_put_cmd (LCD_DISPLAY_ON);
  while (lcd_dt_busy ());
  proto_status (10, STATUS_CONT);
  lcd_put_cmd (LCD_RETURN_HOME);
  while (lcd_dt_busy ());
  proto_status (11, STATUS_CONT);
  lcd_put_cmd (LCD_SET_DDADDR | 0);
  while ((c = pgm_read_byte(s++))) {
    while (lcd_dt_busy ());
    /*PORTB--;*/
    lcd_put_data (c);
  }
  return 0;
}

void
lcd_put_char (uint8_t b)
{
  while (lcd_dt_busy ());
  lcd_put_data (b);
}

void
lcd_home (void)
{
  while (lcd_dt_busy ());
  lcd_put_cmd(LCD_RETURN_HOME);
  while (lcd_dt_busy ());
  lcd_put_cmd(LCD_SET_DDADDR | 0);
}
