/* NMEA parser */
/* Copyright (C) 2010 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>

#include "nmea_hl.h"

/*!\file
 * \ingroup geXPal_lib
 * \brief NMEA parser implementation
 */

/*!\brief size of parser input buffer, including trailing '\0' */
#define NMEA_BUFLEN (82)

/*!\brief store one dimension of coordinate
 */
typedef struct nmea_degree_s {
  uint8_t deg;      /*!<\brief degrees */
  uint8_t min;      /*!<\brief minutes */
  uint16_t minfrac; /*!<\brief 1/1000 of minute */
  char direction;   /*!<\brief North/East/South/West */
} nmea_degree_t;

/*!\brief store GPS timestamp (time of day)
 */
typedef struct nmea_time_s {
  uint8_t hour; /*!<\brief hour (0-23) */
  uint8_t min; /*!<\brief minute (0-59) */
  uint8_t sec; /*!<\brief second (0-59) */
  uint16_t msec; /*!<\brief milliseconds (0-999) */
} nmea_time_t;

/*!\brief store GPS date (day)
 */
typedef struct nmea_date_s {
  uint16_t year; /*!<\brief year */
  uint8_t month; /*!<\brief month (1-12) */
  uint8_t day;   /*!<\brief day (1-31) */
} nmea_date_t;

/*!\brief internal state of parser
 */
struct nmea_parser_s {
  uint8_t bufidx; /*!<\brief offset into buf for next character */
  nmea_time_t time;  /*!<\brief GPS timestamp */
  nmea_date_t date;  /*!<\brief GPS date */
  nmea_degree_t lat; /*!<\brief latitude */
  nmea_degree_t lon; /*!<\brief longitude */
  uint8_t gga_fix;   /*!<\brief type of fix from GGA sentence */
  uint8_t num_sats;  /*!<\brief number of satellites in view */
  struct {
    uint8_t dop;
    uint8_t dopfrac;
  } hdop;         /*!<\brief horizontal dilution of precision */
  struct {
    uint16_t height;
    uint8_t heightfrac;
    char unit;
  } height;       /*!<\brief height above mean sea level */
  struct {
    uint16_t speed;
    uint8_t speedfrac;
    char unit;
  } speed;        /*!<\brief speed over ground */
  struct {
    uint16_t course;
    uint8_t coursefrac;
  } course;       /*!<\brief course over ground */
  nmea_sat_t satellites[MAX_NUM_SATS]; /*!<\brief satellite data */
  uint8_t sat_num_msgs; /*!<\brief number of $GPSV messages */
  char buf[NMEA_BUFLEN]; /*!<\brief input buffer */
};

nmea_parser_t * new_nmea_parser (void)
{
  nmea_parser_t * p = malloc (sizeof (*p));
  uint8_t i;
  p->bufidx = 0;
  p->num_sats = 0;
  p->sat_num_msgs = 0;
  for (i=0; i < MAX_NUM_SATS; i++)
    p->satellites[i].snr = 0xFF;

  return p;
}


static const char GPGGA [] PROGMEM = "$GPGGA,";
static const char GPRMC [] PROGMEM = "$GPRMC,";
static const char GPGSV [] PROGMEM = "$GPGSV,";
static const char GPZDA [] PROGMEM = "$GPZDA,";

/* keep sorted ASCIIbetically, we will use the array in a binary search */
static const prog_char * msgs[] =
  { GPGGA,
    GPGSV,
    GPRMC,
    GPZDA,
    NULL };

/*!
 * The following NMEA sequences have been seen from NL-501:
 *
 * $GPGGA,182401.496,,,,,0,00,,,M,0.0,M,,0000*53
 * $GPGLL,,,,,182401.496,V,N*7F
 * $GPGSA,A,1,,,,,,,,,,,,,,,*1E
 * $GPGSV,3,1,12,20,00,000,,10,00,000,,31,00,000,,27,00,000,*7C
 * $GPGSV,3,2,12,19,00,000,,07,00,000,,04,00,000,,24,00,000,*76
 * $GPGSV,3,3,12,16,00,000,,28,00,000,,26,00,000,,29,00,000,*78
 * $GPRMC,182401.496,V,,,,,,,090210,,,N*42
 * $GPVTG,,T,,M,,N,,K,N*2C
 *
 */

/*!\brief parse NMEA line already in buffer */
static nmea_line_status_t nmea_line_parse (nmea_parser_t * parser)
{
  nmea_parser_t * p = parser;
  int n;
  if (strncmp_P (p->buf, GPGSV, strlen_P (GPGSV)) == 0) {
    uint8_t num_msg;
    nmea_sat_t sats;
    int i; /* index for start of SAT part */
    int8_t k;
    char *s;
    /* Shit. sscanf() cannot handle the case of an empty string at
     * the place to parse in the way we need it (even %[] is limited
     * to _non-empty_ strings).
     * Therefore we parse the fixed part and then for each satellite
     * use sscanf("%d,%n") to split up the message.
     */
    s = p->buf+strlen_P(GPGSV);
    n = sscanf_P (s,
		  PSTR("%1hhu,%1hhu,%2hhu,%n"),
		  p->sat_num_msgs, &num_msg, &p->num_sats, &i
		  );
    printf_P (PSTR("GSV(%d) %s\r\n"), n, s);
    s += i;

    /* the index in GSV starts with 1, compensate by starting the loop
     * variable at -4
     */
    for (k=-4; k < 0; k++) {
      n = sscanf_P(s, PSTR("%2hhu,%2hhu,%3hu,%n"),
		   &sats.id, &sats.elevation, &sats.azimuth, &i);
      s += i;
      if (*s != ',' && *s != '*') {
	n = sscanf_P (s, PSTR("%2hhu%n"), &sats.snr, &i);
	s += i;
      } else {
	sats.snr = 0xFF;
      } 
      if (*s == ',') s++;
      if (num_msg*4+k < MAX_NUM_SATS)
	p->satellites[num_msg*4+k] = sats;
    }
    /* TODO: error checking */
    return NMEA_LINE_SAT;
  } else if (strncmp_P (p->buf, GPGGA, strlen_P (GPGGA)) == 0) {
    n = sscanf_P (p->buf+strlen_P (GPGGA),
		  PSTR("%02hhu%02hhu%02hhu.%03hu,%02hhu%02hhu.%04hu,%c,%03hhu%02hhu.%04hu,%c,%1hhu,%02hhu,%3hhu.%1hhu,%5hd.%01hhu,%c,%*d.%*d,%*c,%*u,%*04d*"),
		  &p->time.hour, &p->time.min, &p->time.sec, &p->time.msec,
		  &p->lat.deg, &p->lat.min, &p->lat.minfrac, &p->lat.direction,
		  &p->lon.deg, &p->lon.min, &p->lon.minfrac, &p->lon.direction,
		  &p->gga_fix, &p->num_sats, &p->hdop.dop, &p->hdop.dopfrac,
		  &p->height.height, &p->height.heightfrac, &p->height.unit);
    printf_P (PSTR("GGA(%d) %s\r\n"), n, p->buf+strlen_P(GPGGA));
    if (n >= 19)
      return NMEA_LINE_FIX;
  } else if (strncmp_P (p->buf, GPRMC, strlen_P (GPRMC)) == 0) {
    char status;
    uint8_t year;
    n = sscanf_P (p->buf+strlen_P (GPRMC),
		  PSTR("%02hhu%02hhu%02hhu.%03hu,%c,%02hhu%02hhu.%04hu,%c,%03hhu%02hhu.%04hu,%c,%hd.%02hhu,%hd.%02hu,%02hhu%02hhu%02hhu"),
		  &p->time.hour, &p->time.min, &p->time.sec, &p->time.msec,
		  &status,
		  &p->lat.deg, &p->lat.min, &p->lat.minfrac, &p->lat.direction,
		  &p->lon.deg, &p->lon.min, &p->lon.minfrac, &p->lon.direction,
		  &p->speed.speed, &p->speed.speedfrac,
		  &p->course.course, &p->course.coursefrac,
		  &p->date.day, &p->date.month, &year );
    printf_P (PSTR("RMC(%d) %s\r\n"), n, p->buf+strlen_P(GPRMC));
    if (n >= 20) {
      /* keep century */
      p->date.year = (p->date.year/100)*100 + year;
      return NMEA_LINE_FIX;
    }
  } else if (strncmp_P (p->buf, GPZDA, strlen_P (GPZDA)) == 0) {
    n = sscanf_P (p->buf+strlen_P (GPZDA),
		  PSTR("%02hhu%02hhu%02hhu,%02hhu,%02hhu,%04hu"),
		  &p->time.hour, &p->time.min, &p->time.sec,
		  &p->date.day, &p->date.month, &p->date.year);
    printf_P (PSTR("ZDA(%d) %s\r\n"), n, p->buf+strlen_P(GPZDA));
    if (n >= 6)
      return NMEA_LINE_FIX;
  }
  return NMEA_LINE_OK;
}

int nmea_format_position (char * buf, size_t sz, nmea_parser_t * parser)
{
  nmea_parser_t * p = parser;
  return snprintf_P (buf, sz,
		     PSTR("%02u%02u%02u.%02u %c%02u\370%02u'%02u\"%c%03u\370%02u'%02u\""),
		     p->time.hour, p->time.min, p->time.sec, p->time.msec/10,
		     p->lat.direction, p->lat.deg, p->lat.min, (p->lat.minfrac*6+500)/1000,
		     p->lon.direction, p->lon.deg, p->lon.min, (p->lon.minfrac*6+500)/1000
		     );
}

nmea_sat_t * nmea_get_satellite (nmea_parser_t *parser, uint8_t idx)
{
  nmea_parser_t * p = parser;
  if (idx < sizeof(p->satellites)/sizeof(p->satellites[0])
      && idx < p->num_sats)
    return &p->satellites[idx];
  else
    return NULL;
}

/*!\brief convert character into hex digit
 *
 * \param h character to convert
 * \return value of character interpreted as hex digit, >= 16 if error
 */
static inline uint8_t hex2int (uint8_t h)
{
  h -= '0'; /* unsigned overflow is deliberate */
  if (h < 10) return h;
  h -= 'A' - '0';
  if (h <= 'F' - 'A') return h + 10;
  h -= 'a' - 'A';
  if (h <= 'f' - 'a') return h + 10;
  return h;
}

/*!\brief check NMEA checksum */
static uint8_t nmea_line_csum_check (nmea_parser_t * parser)
{
  uint8_t csum = 0;
  char * s;
  uint8_t h;
  for (s = parser->buf+1 /*ignore starting '$'*/; *s && *s != '*'; s++) {
    csum ^= *s;
  }
  if (*s++ != '*') return '*';
  h = hex2int(*s++);
  if (h & 0xF0) return h & 0xF0;
  csum ^= h << 4;
  h = hex2int(*s++);
  if (h & 0xF0) return h & 0xF0;
  csum ^= h;
  if (*s != '\r' && *s != '\n') return *s;
  return csum;
}

nmea_line_status_t nmea_recvd (nmea_parser_t * parser, char c)
{
  nmea_parser_t * p = parser; /* for shorter source code */
  if (p->bufidx >= sizeof(p->buf)-1) {
    /* no more space in buffer */
    if (c == '\n') {
      /* line finished -> reset */
      p->bufidx = 0;
      /* but return error indication anyway */
    }
    return NMEA_LINE_OVERFLOW;
  }
  p->buf[p->bufidx++] = c;
  if (c != '\n') {
    /* not finished yet */
    return NMEA_LINE_MORE;
  }

  /* put in terminating NUL byte */
  p->buf[p->bufidx] = '\0';
  /* reset start index */
  p->bufidx = 0;

  /* check checksum */
  if (nmea_line_csum_check (p)) {
    printf_P (PSTR("csum error\r\n"));
    return NMEA_LINE_CKERR;
  }

  /* now we have a complete line with correct checksum
   * and we can start parsing */
  return nmea_line_parse (p);
}
