/* read and parse NMEA messages */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <avr/pgmspace.h>
#include <string.h>
#include <stdio.h>

#include "nmea_reader.h"
#include "serial.h"
#include "events.h"

#define NMEA_BUFLEN 85
static char nmea_buffer[NMEA_BUFLEN];
static uint8_t nmea_bufidx;
static uint8_t nmea_is_checksummed;
static uint8_t nmea_errors;

static const char nmea_gga[] PROGMEM = "$GPGGA";


static void nmea_parse(void);

enum error_code {
  NMEA_BUFFER_OVERFLOW,
  NMEA_CHECKSUM_ERROR,
};

static void
nmea_error(enum error_code c)
{
  nmea_errors |= 1 << c;
}


uint8_t
nmea_init (void)
{
  serial_init (NMEA_SPEED);
  return 0;
}

uint8_t
nmea_checksum (const char * s, uint8_t len)
{
  uint8_t sum;
  for (sum=0; len; s++,len--)
    sum ^= *s;
  return sum;
}

void
nmea_serial_in_handler (uint8_t b)
{
  if (b == NMEA_MSG_END2 && (nmea_buffer[nmea_bufidx-1] == NMEA_MSG_END1
			     || nmea_bufidx == NMEA_BUFLEN)) {
    /* got complete message -> parse it */
    nmea_parse ();
    nmea_bufidx = 0;
    nmea_is_checksummed = 0;
    return;
  }

  if (nmea_bufidx == NMEA_BUFLEN) {
    /* buffer overflow -> ignore char */
    nmea_error(NMEA_BUFFER_OVERFLOW);
    return;
  }

  if (b == NMEA_MSG_STOP) {
    nmea_is_checksummed = nmea_bufidx;
  }

  nmea_buffer[nmea_bufidx++] = b;
}

static uint16_t
convert_hex_digit (char c)
{
  if (c >= '0' && c <= '9') return c-'0';
  else if (c >= 'A' && c <= 'F') return c - 'A' + 10;
  else return 0x100; /* signal error */
}

static uint16_t
hex2_to_int8 (char c1, char c2)
{
  return (convert_hex_digit (c1) << 4) | convert_hex_digit (c2);
}

static struct nmea_ev_data {
  int32_t time;
  int16_t time1000;
  int16_t lati, latf;
  char latsign;
  int16_t loni, lonf;
  char lonsign;
  int16_t quality;
  int16_t numsats;
} nmea_event_data;

void
nmea_parse (void)
{
  uint8_t i,j, l;
  char * s;
  char * s1;
  l = nmea_bufidx;
  if (nmea_is_checksummed) {
    uint16_t cs_nmea;
    i = nmea_is_checksummed;
    cs_nmea = hex2_to_int8(nmea_buffer[i+1], nmea_buffer[i+2]);
    if (cs_nmea && 0xff00) {
      nmea_error (NMEA_CHECKSUM_ERROR);
    } else if ((uint8_t)cs_nmea != nmea_checksum (nmea_buffer,i)) {
      nmea_error (NMEA_CHECKSUM_ERROR);
    }
    l = i;
  }

  if (memcmp_P (nmea_buffer, nmea_gga, strlen_P(nmea_gga)) == 0) {
    struct nmea_ev_data * n = &nmea_event_data;
    /* now it is becoming interesting - we've got a GPS position */
    i = strlen_P (nmea_gga)+1; /* position at first data field */
    s = nmea_buffer+i;
    if (sscanf_P (s, "%6i.%h3i,%hi.%hi,%c,%hi.%hi,%c,%h1i,%h2i,",
		  &n->time, &n->time1000,
		  &n->lati, &n->latf, &n->latsign,
		  &n->loni, &n->lonf, &n->lonsign,
		  &n->quality, &n->numsats
		  ) == 10) {
      put_nmea_event (n->time, n->lati, n->latf, n->latsign == 'N' ? '+' : '-',
		      n->loni, n->lonf, n->lonsign == 'E' ? '+' : '-',
		      0, 0, 0);
    }
  }
}
