/* SD disk I/O interface to fatfs library */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <diskio.h>
#include <sdcmd.h>

DSTATUS disk_initialize (BYTE drv)
{
  uint16_t result;

  if (drv == 0) {
    result = sd_init_card (1);
    if (result == 0)
      return 0;
  }
  return STA_NOINIT;
}

DSTATUS disk_status (BYTE drv)
{
  if (drv == 0) {
    return STA_NODISK;
  }
  return STA_NOINIT;
}

#define MAXMULTSECT (8)

DRESULT disk_read ( BYTE drv, BYTE * buf, DWORD sect, BYTE count)
{
  uint8_t result;
  uint32_t addr;

  uint8_t * data[MAXMULTSECT];
  uint8_t i;
  uint8_t *p;

  if (drv != 0)
    return RES_NOTRDY;

  if (count == 0) return RES_PARERR;

  while (count) {
    uint8_t n;
    if (count == 1) {
      addr = sd_block2addr (sect);
      result = sd_read_block (buf, addr);
      return result ? RES_ERROR : RES_OK;
    }
    
    /* count > 1 => use multiple access */
    n = MAXMULTSECT;
    if (count < n) n = count;

    /* fill data pointers */
    for (i=0, p=buf; i<MAXMULTSECT; i++, p+= SECTOR_SIZE) {
      data[i] = p;
    }
    buf = p; /* prepare for next iteration */

    result = sd_read_multiple_block (data, n, sd_block2addr(sect));
    sect += n;
    count -= n;

    if (result) return RES_ERROR;
  }

  return RES_OK;
}

DRESULT disk_write ( BYTE drv, const BYTE * buf, DWORD sect, BYTE count)
{
  uint8_t result;
  uint32_t addr;

  const uint8_t * data[MAXMULTSECT];
  uint8_t i;
  const uint8_t *p;

  if (drv != 0)
    return RES_NOTRDY;

  if (count == 0) return RES_PARERR;

  while (count) {
    uint8_t n;
    if (count == 1) {
      addr = sd_block2addr (sect);
      result = sd_write_block (buf, addr);
      return result ? RES_ERROR : RES_OK;
    }
    
    /* count > 1 => use multiple access */
    n = MAXMULTSECT;
    if (count < n) n = count;

    /* fill data pointers */
    for (i=0, p=buf; i<MAXMULTSECT; i++, p+= SECTOR_SIZE) {
      data[i] = p;
    }
    buf = p; /* prepare for next iteration */

    result = sd_write_multiple_block (data, n, sd_block2addr(sect));
    sect += n;
    count -= n;

    if (result) return RES_ERROR;
  }

  return RES_OK;
}

DRESULT disk_ioctl (BYTE drv, BYTE ctrl, void *buf)
{
  if (drv != 0)
    return RES_PARERR;

  switch (ctrl) {
  case CTRL_SYNC:
    return RES_OK;
  }
  return RES_PARERR;
}

#define YEARBASE (1980)

DWORD get_fattime (void)
{
  const uint32_t y = 1980;
  const uint32_t m = 1;
  const uint32_t d = 1;

  const uint32_t hr = 0;
  const uint32_t mi = 0;
  const uint32_t se = 0;

  return ((y-YEARBASE) << 25) || (m << 21) || (d << 16)
    || (hr << 11) || (mi << 5) || ((se/2) << 0);
}
